# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.


"""
Test case for module Worker
"""

import errno
import os
import tempfile

from tests import TestCleaner, common
from tests.TestFileUtilities import _open_blocking_handle
import bleachbit
from bleachbit import CLI, Command
from bleachbit.Action import ActionProvider
from bleachbit.Cleaner import backends
from bleachbit.Worker import Worker

if bleachbit.IS_WINDOWS:
    import win32con
    import win32file


class AccessDeniedActionAction(ActionProvider):
    action_key = 'access.denied'

    def __init__(self, action_element):
        self.pathname = action_element.getAttribute('path')

    def get_commands(self):
        # access denied, should fail and continue
        def accessdenied():
            raise OSError(
                errno.EACCES, 'Permission denied: c:\\access\\denied', 'c:\\access\\denied')
        yield Command.Function(None, accessdenied, 'Test access denied')

        # real file, should succeed
        yield Command.Delete(self.pathname)


class DoesNotExistAction(ActionProvider):
    action_key = 'does.not.exist'

    def __init__(self, action_element):
        self.pathname = action_element.getAttribute('path')

    def get_commands(self):
        # non-existent file, should fail and continue
        if bleachbit.IS_WINDOWS:
            yield Command.Delete(r"c:\does\not\exist")
        else:
            yield Command.Delete("/does/not/exist")

        # real file, should succeed
        yield Command.Delete(self.pathname)


class FunctionGeneratorAction(ActionProvider):
    action_key = 'function.generator'

    def __init__(self, action_element):
        self.pathname = action_element.getAttribute('path')

    def get_commands(self):
        # function generator without path, should succeed
        def funcgenerator():
            yield 10
        yield Command.Function(None, funcgenerator, 'funcgenerator')

        # real file, should succeed
        yield Command.Delete(self.pathname)


class FunctionPathAction(ActionProvider):
    action_key = 'function.path'

    def __init__(self, action_element):
        self.pathname = action_element.getAttribute('path')

    def get_commands(self):
        # function with path, should succeed
        def pathfunc(_path):
            pass
        # self.pathname must exist because it checks the file size
        yield Command.Function(self.pathname, pathfunc, 'pathfunc')

        # real file, should succeed
        yield Command.Delete(self.pathname)


class InvalidEncodingAction(ActionProvider):
    action_key = 'invalid.encoding'

    def __init__(self, action_element):
        self.pathname = action_element.getAttribute('path')

    def get_commands(self):
        # file with invalid encoding
        (fd, filename) = tempfile.mkstemp('invalid-encoding-\xe4\xf6\xfc~')
        os.close(fd)
        yield Command.Delete(filename)

        # real file, should succeed
        yield Command.Delete(self.pathname)


class FunctionPlainAction(ActionProvider):
    action_key = 'function.plain'

    def __init__(self, action_element):
        self.pathname = action_element.getAttribute('path')

    def get_commands(self):
        # plain function without path, should succeed
        def intfunc():
            return int(5)
        yield Command.Function(None, intfunc, 'intfunc')

        # real file, should succeed
        yield Command.Delete(self.pathname)


class LockedAction(ActionProvider):

    action_key = 'locked'

    def __init__(self, action_element):
        self.pathname = action_element.getAttribute('path')

    def get_commands(self):
        # Open the file with a blocking handle that allows read/write but blocks delete.
        # This causes delete() to truncate and mark for deletion. Windows only.
        share_mode = win32con.FILE_SHARE_READ | win32con.FILE_SHARE_WRITE
        handle = _open_blocking_handle(self.pathname, share_mode)
        from bleachbit.FileUtilities import getsize
        # Without admin privileges, this delete fails.
        yield Command.Delete(self.pathname)
        assert (os.path.exists(self.pathname))
        fsize = getsize(self.pathname)
        if not fsize == 0:  # File should be truncated to 0 bytes
            raise RuntimeError('Locked file has size %dB (not 0B)' % fsize)
        win32file.CloseHandle(handle)

        # Now that the file is not locked, admin privileges
        # are not required to delete it.
        yield Command.Delete(self.pathname)


class RuntimeErrorAction(ActionProvider):

    action_key = 'runtime'

    def __init__(self, action_element):
        self.pathname = action_element.getAttribute('path')

    def get_commands(self):
        # runtime exception, should fail and continue
        def runtime():
            raise RuntimeError('This is a test exception')

        yield Command.Function(None, runtime, 'Test runtime exception')

        # real file, should succeed
        yield Command.Delete(self.pathname)


class TruncateTestAction(ActionProvider):
    action_key = 'truncate.test'

    def __init__(self, action_element):
        self.pathname = action_element.getAttribute('path')

    def get_commands(self):
        # truncate real file
        yield Command.Truncate(self.pathname)

        # real file, should succeed
        yield Command.Delete(self.pathname)


class WorkerTestCase(common.BleachbitTestCase):

    """Test case for module Worker"""

    def action_test_helper(self, command, special_expected, errors_expected,
                           bytes_expected_posix, count_deleted_posix,
                           bytes_expected_nt, count_deleted_nt):
        ui = CLI.CliCallback()
        (fd, filename) = tempfile.mkstemp(
            prefix='bleachbit-test-worker', dir=self.tempdir)
        os.write(fd, b'123')
        os.close(fd)
        self.assertExists(filename)
        astr = '<action command="%s" path="%s"/>' % (command, filename)
        cleaner = TestCleaner.action_to_cleaner(astr)
        backends['test'] = cleaner
        operations = {'test': ['option1']}
        worker = Worker(ui, True, operations)
        run = worker.run()
        while next(run):
            pass
        del backends['test']
        self.assertNotExists(filename, "Path still exists '%s'" % filename)
        self.assertEqual(worker.total_special, special_expected,
                         'For command %s expecting %s special operations but observed %d'
                         % (command, special_expected, worker.total_special))
        self.assertEqual(worker.total_errors, errors_expected,
                         'For command %s expecting %d errors but observed %d'
                         % (command, errors_expected, worker.total_errors))
        if bleachbit.IS_POSIX:
            self.assertEqual(worker.total_bytes, bytes_expected_posix)
            self.assertEqual(worker.total_deleted, count_deleted_posix)
        elif bleachbit.IS_WINDOWS:
            self.assertEqual(worker.total_bytes, bytes_expected_nt)
            self.assertEqual(worker.total_deleted, count_deleted_nt)

    def test_AccessDenied(self):
        """Test Worker using Action.AccessDeniedAction"""
        with self.assertLogs(level='ERROR') as log_context:
            self.action_test_helper('access.denied', 0, 1, 4096, 1, 3, 1)
            if bleachbit.IS_WINDOWS:
                self.assertIn('Access denied: c:\\access\\denied',
                              log_context.output[0])
                self.assertNotIn('\\\\', log_context.output[0])

    def test_DoesNotExist(self):
        """Test Worker using Action.DoesNotExistAction"""
        with self.assertLogs(level='ERROR') as log_context:
            self.action_test_helper('does.not.exist', 0, 1, 4096, 1, 3, 1)
            if bleachbit.IS_WINDOWS:
                # Make sure there is a simple messages without double backslashes.
                self.assertIn(
                    'File not found: c:\\does\\not\\exist', log_context.output[0])
                self.assertNotIn('\\\\', log_context.output[0])

    def test_FunctionGenerator(self):
        """Test Worker using Action.FunctionGenerator"""
        self.action_test_helper('function.generator', 1,
                                0, 4096 + 10, 1, 3 + 10, 1)

    def test_FunctionPath(self):
        """Test Worker using Action.FunctionPathAction"""
        self.action_test_helper('function.path', 1, 0, 4096, 1, 3, 1)

    def test_FunctionPlain(self):
        """Test Worker using Action.FunctionPlainAction"""
        self.action_test_helper('function.plain', 1, 0, 4096 + 5, 1, 3 + 5, 1)

    def test_InvalidEncoding(self):
        """Test Worker using Action.InvalidEncodingAction"""
        self.action_test_helper('invalid.encoding', 0, 0, 4096, 2, 3, 2)

    @common.skipUnlessWindows
    def test_Locked(self):
        """Test Worker using Action.LockedAction"""
        from win32com.shell import shell
        if shell.IsUserAnAdmin():
            # If an admin, the first attempt, with locking, will mark for
            # delete (3 bytes), and the second attempt will actually delete
            # it (0 bytes, already truncated).
            errors_expected = 0
            bytes_expected = 3 + 0
            total_deleted = 2
        else:
            # If not an admin, the first attempt (locked) truncates (to 0B) and fails,
            # then the second attempt (unlocked) deletes the 0-byte file.
            errors_expected = 1
            bytes_expected = 0
            total_deleted = 1
        self.action_test_helper(
            'locked', 0, errors_expected, None, None, bytes_expected, total_deleted)

    def test_RuntimeError(self):
        """Test Worker using Action.RuntimeErrorAction. It is normal to see a traceback.

        The Worker module handles these differently than
        access denied exceptions
        """
        self.action_test_helper('runtime', 0, 1, 4096, 1, 3, 1)

    def test_Truncate(self):
        """Test Worker using Action.TruncateTestAction
        """
        self.action_test_helper('truncate.test', 0, 0, 4096, 2, 3, 2)

    def test_deep_scan(self):
        """Test for deep scan"""

        # load cleaners from XML
        import bleachbit.CleanerML
        list(bleachbit.CleanerML.load_cleaners())

        # DeepScan itself is tested elsewhere, so replace it here
        import bleachbit.DeepScan
        SaveDeepScan = bleachbit.DeepScan.DeepScan
        self.scanned = 0
        parent = self

        class MyDeepScan:
            def __init__(self, searches):
                for (path, searches) in searches.items():
                    parent.assertEqual(path, os.path.expanduser('~'))
                    for s in searches:
                        parent.assertIn(
                            s.regex, ['^Thumbs\\.db$', '^Thumbs\\.db:encryptable$'])

            def scan(self):
                parent.scanned += 1
                yield True

        bleachbit.DeepScan.DeepScan = MyDeepScan

        # test
        operations = {'deepscan': ['thumbs_db']}
        ui = CLI.CliCallback()
        worker = Worker(ui, False, operations).run()
        while next(worker):
            pass
        self.assertEqual(1, self.scanned)

        # clean up
        bleachbit.DeepScan.DeepScan = SaveDeepScan

    def test_multiple_options(self):
        """Test one cleaner with two options"""
        ui = CLI.CliCallback()
        filename1 = self.mkstemp(prefix='bleachbit-test-worker')
        filename2 = self.mkstemp(prefix='bleachbit-test-worker')

        astr1 = '<action command="delete" search="file" path="%s"/>' % filename1
        astr2 = '<action command="delete" search="file" path="%s"/>' % filename2
        cleaner = TestCleaner.actions_to_cleaner([astr1, astr2])
        backends['test'] = cleaner
        operations = {'test': ['option1', 'option2']}
        worker = Worker(ui, True, operations)
        run = worker.run()
        while next(run):
            pass
        del backends['test']
        self.assertNotExists(filename1)
        self.assertNotExists(filename2)
        self.assertEqual(worker.total_special, 0)
        self.assertEqual(worker.total_errors, 0)
        self.assertEqual(worker.total_deleted, 2)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.


"""
Test case for module SystemInformation
"""

import os
import re
from unittest import mock


from tests import common
from bleachbit import IS_POSIX, FS_SCAN_RE_FLAGS
from bleachbit.SystemInformation import _get_home_dirs_to_anonymize, get_system_information, get_version, anonymize_system_information


class SystemInformationTestCase(common.BleachbitTestCase):
    """Test Case for module SystemInformation"""

    def test_get_system_information(self):
        """Test get_system_information"""
        # at least it does not crash
        ret = get_system_information()
        self.assertIsString(ret)

    @common.skipUnlessWindows
    def test_get_system_information_invalid_unicode_userprofile(self):
        invalid_userprofile = 'C:\\Users\\invalid\ud803'
        original_getenv = os.getenv

        def getenv(name, default=None):
            if name == 'USERPROFILE':
                return invalid_userprofile
            return original_getenv(name, default)

        with mock.patch('bleachbit.SystemInformation.os.getenv', side_effect=getenv):
            ret = get_system_information()

        encoded = ret.encode('utf-8')
        self.assertIn(b'C:\\Users\\invalid\\ud803', encoded)

    def test_get_version(self):
        """Test get_version"""
        ret = get_version()
        self.assertIsString(ret)
        # check it is in a.b.c.d or a.b.c format
        self.assertRegex(ret, r'^\d+\.\d+\.\d+(\.\d+)?$')
        ret = get_version(four_parts=True)
        self.assertIsString(ret)
        # check it is in a.b.c.d format
        self.assertRegex(ret, r'^\d+\.\d+\.\d+\.\d+$')

    @common.skipIfWindows
    def test_anonymize_system_information_posix_non_root(self):
        """Test anonymization on POSIX with a non-root home directory."""
        home_dir = os.path.expanduser('~')
        if home_dir in ('/root', '/'):
            self.skipTest('requires a non-root POSIX home directory')

        test_input = f"""BleachBit version = 5.1.0
local_cleaners_dir = {home_dir}/.local/share/bleachbit/cleaners
options_dir = {home_dir}/.config/bleachbit
os.getenv(LOGNAME) = testuser
os.getenv(USER) = testuser
os.getenv(SUDO_UID) = 1000"""

        result = anonymize_system_information(test_input)

        self.assertIn(
            'local_cleaners_dir = ~/.local/share/bleachbit/cleaners', result)
        self.assertIn('options_dir = ~/.config/bleachbit', result)
        self.assertNotIn(home_dir, result)
        self.assertNotIn(home_dir.lower(), result.lower())
        self.assertIn('os.getenv(LOGNAME) = *non-root*', result)
        self.assertIn('os.getenv(USER) = *non-root*', result)
        self.assertIn('os.getenv(SUDO_UID) = 1000', result)

    @common.skipUnlessWindows
    def test_anonymize_system_information_windows(self):
        """Test anonymization on Windows with environment variables."""
        home_dir = os.path.expanduser('~')
        test_input = f"""os.getenv(APPDATA) = {home_dir.upper()}\\AppData\\Roaming
os.getenv(USERPROFILE) = {home_dir.lower()}
os.getenv(LocalAppData) = {home_dir}\\AppData\\Local"""

        result = anonymize_system_information(test_input)

        self.assertIn(
            'os.getenv(APPDATA) = %userprofile%\\AppData\\Roaming', result)
        self.assertIn(
            'os.getenv(USERPROFILE) = %userprofile%', result)
        self.assertIn(
            'os.getenv(LocalAppData) = %userprofile%\\AppData\\Local', result)
        self.assertIsNone(re.search(re.escape(home_dir),
                          result, flags=FS_SCAN_RE_FLAGS))

    @common.skipUnlessWindows
    def test_anonymize_system_information_windows_short_path(self):
        """Test anonymization handles Windows short (8.3) paths."""
        # Create a directory with Unicode characters that will have a short path
        username = '我可以喝漂白剂而不伤及自身'
        unicode_dir = self.mkdir(username)

        try:
            import win32api
            short_path = win32api.GetShortPathName(unicode_dir)
        except (ImportError, OSError):
            self.skipTest('win32api not available or short path not supported')

        if short_path == unicode_dir:
            self.skipTest('Short path generation not enabled on this system')

        self.assertTrue(os.path.samefile(unicode_dir, short_path))

        with mock.patch('bleachbit.SystemInformation.os.path.expanduser', return_value=unicode_dir):
            from bleachbit.SystemInformation import _get_home_dirs_to_anonymize
            home_dirs = _get_home_dirs_to_anonymize()

        self.assertIn(short_path, home_dirs)
        self.assertTrue(os.path.samefile(home_dirs[0], home_dirs[1]))

        test_input = f"os.getenv(TMP) = {short_path}\\AppData\\Local\\Temp"
        with mock.patch('bleachbit.SystemInformation._get_home_dirs_to_anonymize',
                        return_value=[unicode_dir, short_path]):
            anonymized = anonymize_system_information(test_input)

        self.assertIn('%userprofile%', anonymized)
        self.assertNotIn(username, anonymized)
        self.assertNotIn(short_path, anonymized)
        self.assertNotIn(unicode_dir, anonymized)

    def test_anonymize_system_information_preserves_non_sensitive_data(self):
        """Test that non-sensitive data remains unchanged during anonymization."""
        test_unchanged = """BleachBit version = 5.1.0
gi.version = 3.55.0
GTK version = 3.24.51"""

        result = anonymize_system_information(test_unchanged)
        self.assertEqual(test_unchanged, result)

    def test_anonymize_system_information_preserves_root(self):
        """Root username should not be mislabeled as non-root."""
        test_input = """BleachBit version = 5.1.0
local_cleaners_dir = /root/.local/share/bleachbit/cleaners
options_dir = /root/.config/bleachbit
personal_cleaners_dir = /root/.config/bleachbit/cleaners
os.getenv(LOGNAME) = root
os.getenv(USER) = root
os.getenv(SUDO_UID) = 1000
os.path.expanduser(~") = /root"""
        result = anonymize_system_information(test_input)
        self.assertEqual(test_input, result)

    def test_anonymize_system_information_masks_real_home_dir(self):
        """Preserve /root while anonymizing the real non-root home directory."""
        # Mixing /root and user-specific paths is a normal scenario when
        # running from source `sudo python3 bleachbit.py`.
        test_input = """personal_cleaners_dir = /root/.config/bleachbit/cleaners
system_cleaners_dir = /home/regularuser/software/bleachbit/cleaners
__file__ = /home/regularuser/software/bleachbit/bleachbit/SystemInformation.py"""
        home_token = '~' if IS_POSIX else '%userprofile%'

        with mock.patch('bleachbit.SystemInformation._get_home_dirs_to_anonymize',
                        return_value=['/home/regularuser']):
            result = anonymize_system_information(test_input)

        self.assertIn(
            f'system_cleaners_dir = {home_token}/software/bleachbit/cleaners', result)
        self.assertIn(
            f'__file__ = {home_token}/software/bleachbit/bleachbit/SystemInformation.py', result)
        self.assertNotIn('/home/regularuser', result)

    def test_anonymize_preserves_structure(self):
        """Test that anonymization preserves the structure of system information"""
        # This is the only anonymization test with the real get_system_information().
        original = get_system_information()
        anonymized = anonymize_system_information(original)

        # Both should have the same number of lines
        self.assertEqual(len(original.split('\n')),
                         len(anonymized.split('\n')))

        # Both should have the same number of equal signs (key-value delimiters)
        self.assertEqual(original.count(' = '), anonymized.count(' = '))

        # Both should be non-empty strings
        self.assertIsString(original)
        self.assertIsString(anonymized)
        self.assertTrue(len(anonymized) > 0)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.


"""
Test case for module GUI
"""

import os
import sys
import unittest
import time
import types
import warnings
from unittest import mock

import bleachbit
from bleachbit.Cleaner import backends
from bleachbit.GtkShim import HAVE_GTK, Gdk, Gio, GLib, GObject, Gtk
from bleachbit.Language import get_text as _
from bleachbit.Options import options
from tests import common

if HAVE_GTK:
    from bleachbit.GuiUtil import (clear_clipboard, get_font_size_from_name,
                                   get_window_info)
    from bleachbit.GuiTreeModels import TreeDisplayModel

bleachbit.online_update_notification_enabled = False


@unittest.skipUnless(HAVE_GTK, 'requires GTK+ module and a display environment')
class GUITestCase(common.BleachbitTestCase):
    app = None
    options_get_tree = options.get_tree
    _NEW_CLEANER_ID, _NEW_OPTION_ID = 'test_run_operations', 'test1'

    """Test case for module GUI"""
    @classmethod
    def setUpClass(cls):
        from bleachbit.GuiApplication import Bleachbit
        cls.app = Bleachbit(auto_exit=False, uac=False)
        cls._lang_env = common.set_temporary_env('LANGUAGE', 'en')
        cls._lang_env.__enter__()
        # reminder: the set up in the parent class creates a clean
        # configuration file.
        super(GUITestCase, GUITestCase).setUpClass()
        options.get_tree = types.MethodType(
            lambda self, parent, child: False, options)
        options.set('font_check_completed', True)
        try:
            cls.app.register()
        except GLib.GError as e:
            if "already exported" in str(e):
                cls.app = Gio.Application.get_default()
            else:
                raise
        with common.capture_glib_exceptions() as glib_errors:
            cls.app.activate()
            cls.refresh_gui()
        if glib_errors:
            _exc_type, exc_value, exc_tb = glib_errors[0]
            raise exc_value.with_traceback(exc_tb)

    @classmethod
    def tearDownClass(cls):
        super(GUITestCase, GUITestCase).tearDownClass()
        options.get_tree = cls.options_get_tree
        cls._lang_env.__exit__(None, None, None)
        window = cls.get_window()
        if window:
            window.destroy()
            cls.clear_window()

    @classmethod
    def get_window(cls):
        return getattr(cls.app, '_window', None)

    @classmethod
    def clear_window(cls):
        if cls.app:
            setattr(cls.app, '_window', None)

    @classmethod
    def refresh_gui(cls, delay=0):
        while Gtk.events_pending():
            if sys.version_info >= (3, 14):
                with warnings.catch_warnings():
                    warnings.simplefilter("error")
                    warnings.filterwarnings(
                        "ignore",
                        message=".*asyncio.AbstractEventLoopPolicy.*",
                        category=DeprecationWarning
                    )
                    warnings.filterwarnings(
                        "ignore",
                        message=".*asyncio.get_event_loop_policy.*",
                        category=DeprecationWarning
                    )
                    Gtk.main_iteration_do(blocking=False)
            else:
                Gtk.main_iteration_do(blocking=False)
        time.sleep(delay)

    @classmethod
    def print_widget(cls, widget, indent=0):
        print('{}{}'.format(' ' * indent, widget))
        if isinstance(widget, Gtk.Container):
            for c in widget.get_children():
                cls.print_widget(c, indent + 2)

    @classmethod
    def find_widget(cls, widget, widget_class, widget_label=None):
        if isinstance(widget, widget_class):
            if widget_label is None or widget.get_label() == widget_label:
                return widget
        if isinstance(widget, Gtk.Container):
            for c in widget.get_children():
                b = cls.find_widget(c, widget_class, widget_label)
                if b is not None:
                    return b
        return None

    @classmethod
    def get_iter(cls, model, cleaner):
        it = model.get_iter(Gtk.TreePath(0))
        while it:
            if model[it][2] == cleaner:
                return model.iter_children(it)
            it = model.iter_next(it)
        return None

    @classmethod
    def find_option(cls, model, cleaner, option):
        it = cls.get_iter(model, cleaner)
        # cls.assertIsNotNone(cls, it)
        while it:
            if model[it][2] == option:
                return it
            it = model.iter_next(it)
        return None

    def _put_checkmark_on_cleaner(self, gui, cleaner_id, option_id):
        model = gui.view.get_model()
        tree = self.find_widget(gui, Gtk.TreeView)
        self.assertIsNotNone(tree)
        it = self.find_option(model, cleaner_id, option_id)
        self.assertIsNotNone(it)
        tree.scroll_to_cell(model.get_path(it), None, False, 0, 0)
        model[model.iter_parent(it)][1] = True
        model[it][1] = True
        self.refresh_gui()

    def click_button(self, dialog, label):
        b = self.find_widget(dialog, Gtk.Button, label)
        self.assertIsNotNone(b)
        b.clicked()
        self.refresh_gui()

    def wait_until(self, condition, timeout=5):
        """
        Wait until a condition is met or timeout is reached.

        Args:
            condition: A callable that returns a boolean
            timeout: Maximum time to wait in seconds

        Returns:
            True if condition is met, False if timeout is reached
        """
        deadline = time.time() + timeout
        while time.time() < deadline:
            self.refresh_gui(0.05)
            if condition():
                return True
        self.refresh_gui()
        return condition()

    def test_append_text(self):
        """Test append_text handles special strings"""
        gui = self.get_window()
        for test_str in common.SPECIAL_TEST_STRINGS:
            gui.append_text(test_str + "\n")
            gui.append_text(f"prefix{test_str}suffix\n")
            gui.append_text(test_str + "\n", tag='error')
            gui.append_text(f"Deleting file: /path/to/file_{test_str}.txt\n")
        self.refresh_gui()

    def test_get_window_info(self):
        """Test get_window_info"""
        gui = self.get_window()
        geo = get_window_info(gui)
        self.assertGreaterEqual(geo.x, 0)
        self.assertGreaterEqual(geo.y, 0)
        self.assertGreaterEqual(geo.width, 0)
        self.assertGreaterEqual(geo.height, 0)
        self.assertIsInstance(geo.monitor_model, str)
        self.assertGreater(len(geo.monitor_model), 0)
        self.assertNotEqual(geo.monitor_model, "(error)")

    def test_GUI(self):
        """Unit test for class GUI"""
        # there should be no crashes
        # app.do_startup()
        # pp.do_activate()                            Build a unit test that that does this
        gui = self.get_window()
        gui.update_progress_bar(0.0)
        gui.update_progress_bar(1.0)
        gui.update_progress_bar("status")

    def test_get_font_size_from_name(self):
        """Test get_font_size_from_name()"""
        tests = (("Sans 12", 12),
                 ("Sans 10.5", 10),
                 ("Sans 0", None),
                 ("Sans -1", None),
                 ("Sans", None),
                 ("", None),
                 (None, None))
        for font_name, expected in tests:
            self.assertEqual(get_font_size_from_name(
                font_name), expected, f"Font name '{font_name}' should return {expected}")

    def test_preferences(self):
        """Opens the preferences dialog and closes it"""

        # show preferences dialog
        pref = self.app.get_preferences_dialog()
        pref.dialog.show_all()
        self.refresh_gui()

        # click close button
        self.click_button(pref.dialog, Gtk.STOCK_CLOSE)

        # destroy
        pref.dialog.destroy()

    def test_preferences_cookies_page(self):
        """Opens the preferences dialog and navigates to cookies page"""
        pref = self.app.get_preferences_dialog()
        pref.dialog.show_all()
        self.refresh_gui()

        # Navigate to cookies page - this will trigger cookie manager loading
        pref.select_page('cookies')
        self.refresh_gui()

        # click close button
        self.click_button(pref.dialog, Gtk.STOCK_CLOSE)
        pref.dialog.destroy()

    def test_system_information(self):
        """Opens the system information dialog and closes it"""
        dialog, _txt, _txtbuffer = self.app.get_system_information_dialog()
        dialog.show_all()
        self.refresh_gui()

        # click close button
        self.click_button(dialog, Gtk.STOCK_CLOSE)

        # destroy
        dialog.destroy()

    def test_about(self):
        """Opens the about dialog and closes it"""
        about = self.app.get_about_dialog()
        about.show_all()
        self.refresh_gui()

        # destroy
        about.destroy()

    def test_chaff(self):
        """Minimal test of the chaff dialog"""
        from bleachbit import Chaff, GuiChaff
        # common.py patches the download directory, so have_models() will return False.
        if not Chaff.download_models():
            self.skipTest('Unable to download chaff models for GUI test')
        gui = self.get_window()
        cd = GuiChaff.ChaffDialog(gui)
        cd.show_all()
        # Trigger missing-folder branch (no destination chosen)
        cd.choose_folder_button.unselect_all()
        self.refresh_gui()
        cd.on_make_files(None)
        self.refresh_gui()
        self.assertTrue(cd.infobar.get_visible())
        chaff_dst_dir = os.path.join(self.tempdir, 'chaff_dst')
        os.mkdir(chaff_dst_dir)
        cd.choose_folder_button.set_filename(chaff_dst_dir)
        cd.when_finished_combo.set_active(1)  # do not delete after generation
        cd.stop_value_spin.set_value(10)
        self.refresh_gui()
        self.click_button(cd, _("Make files"))
        self.assertIsNotNone(
            cd.thread, 'Chaff generation thread did not start')
        cd.thread.join()
        self.refresh_gui()
        self.assertEqual(len(os.listdir(chaff_dst_dir)), 10)

    def test_cookie_manager_bulk_actions_wait_for_loading(self):
        """Bulk cookie actions should not save until loading finishes"""
        from bleachbit import GuiCookie

        class FakeThread:
            def __init__(self, target=None, daemon=None):
                self.target = target
                self.daemon = daemon

            def start(self):
                return None

        with mock.patch('bleachbit.GuiCookie.threading.Thread', FakeThread):
            pane = GuiCookie.CookieManagerPane()

        self.assertTrue(pane._is_loading)
        self.assertFalse(pane.select_all_btn.get_sensitive())
        self.assertFalse(pane.deselect_all_btn.get_sensitive())

        with mock.patch.object(pane, 'save_changes') as save_changes:
            pane.on_select_all_clicked(None)
            pane.on_deselect_all_clicked(None)
            save_changes.assert_not_called()

        pane._finish_populate(['example.com'])
        self.refresh_gui()

        self.assertFalse(pane._is_loading)
        self.assertTrue(pane.select_all_btn.get_sensitive())
        self.assertTrue(pane.deselect_all_btn.get_sensitive())

        with mock.patch.object(pane, 'save_changes') as save_changes:
            pane.on_select_all_clicked(None)
            save_changes.assert_called_once_with()
        self.assertTrue(all(row[0] for row in pane.cookie_store))
        pane.destroy()

    def test_preview(self):
        """Select cleaner option and clicks preview button"""
        gui = self.get_window()
        self.refresh_gui()
        self._put_checkmark_on_cleaner(gui, 'system', 'tmp')
        self.refresh_gui()
        self.click_button(gui, _("Preview"))
        self.refresh_gui()

    def test_notify(self):
        """Test a pop-up notification"""
        from bleachbit.GuiUtil import notify
        notify('This is a test notification')
        time.sleep(1)  # Pause for 1 second

    @mock.patch('bleachbit.GuiBasic.delete_confirmation_dialog')
    def test_confirm_delete(self, mock_delete_confirmation_dialog):
        gui = self.get_window()
        options.set('expert_mode', True)
        for new_delete_confirmation in [True, False]:
            options.set('delete_confirmation',
                        new_delete_confirmation)
            gui._confirm_delete(False, False)

        # We should have a single call to delete_confirmation_dialog
        # only when delete_confirmation option is True.
        mock_delete_confirmation_dialog.assert_called_once()

    def test_shred_paths(self):
        dirname = self.mkdtemp(prefix='bleachbit-test-shred_paths')
        test_files_dirs = [
            self.mkstemp(prefix="somefile", dir=dirname),
            dirname,
            self.mkstemp(prefix="somefile")
        ]

        for obj in test_files_dirs:
            self.assertExists(obj)

        gui = self.get_window()
        self.refresh_gui()

        with mock.patch('bleachbit.GuiWindow.GUI._confirm_delete', return_value=True):
            self.assertTrue(gui._confirm_delete(False, False))
            gui.shred_paths(test_files_dirs)

        self.refresh_gui()

        for obj in test_files_dirs:
            self.assertNotExists(obj)

    def test_shred_paths_cancel_cleans_temporary_backend(self):
        """Test that shred_paths with cancel cleans up the temporary backend"""
        test_file = self.write_file('shred-cancel')
        gui = self.get_window()
        options.set('delete_confirmation', True)
        self.refresh_gui()

        try:
            with mock.patch.object(gui, '_confirm_delete', return_value=False):
                gui.shred_paths([test_file])

            self.assertTrue(self.wait_until(lambda: '_gui' not in backends))
            self.assertExists(test_file)
        finally:
            options.set('delete_confirmation', False)
            backends.pop('_gui', None)

    def test_shred_paths_cancel_after_preview_cleans_temporary_backend(self):
        """Test cancel cleans up _gui even when the preview finishes during the dialog.

        In real usage the confirmation dialog runs a nested GTK main loop, so
        the preview worker can finish and call worker_done() before the user
        clicks Cancel.  This test simulates that by processing GUI events
        inside the mocked _confirm_delete before returning False.
        """
        test_file = self.write_file('shred-cancel-after-preview')
        gui = self.get_window()
        options.set('delete_confirmation', True)
        self.refresh_gui()

        def fake_confirm_delete(*_args, **_kwargs):
            # Simulate the dialog processing idle events, allowing the
            # preview worker to finish before the user cancels.
            self.wait_until(lambda: '_gui' not in backends
                            or gui._gui_cleaner_cleanup_pending is None)
            self.refresh_gui()
            return False

        try:
            with mock.patch.object(gui, '_confirm_delete',
                                   side_effect=fake_confirm_delete):
                gui.shred_paths([test_file])

            # _gui must be gone even though the preview finished during the
            # dialog (before _confirm_delete returned).
            self.assertNotIn('_gui', backends)
            self.assertExists(test_file)
        finally:
            options.set('delete_confirmation', False)
            backends.pop('_gui', None)

    def test_shred_paths_clears_clipboard_mock(self):
        """Test that shred_paths with should_clear_clipboard=True clears the clipboard"""
        test_file = self.write_file('shred-me-via-clipboard')
        gui = self.get_window()
        self.refresh_gui()

        with mock.patch('bleachbit.GuiWindow.clear_clipboard') as mock_clear_clipboard:
            with mock.patch.object(gui, '_confirm_delete', return_value=True):
                gui.shred_paths([test_file], should_clear_clipboard=True)

        self.refresh_gui()

        mock_clear_clipboard.assert_called_once()
        self.assertNotExists(test_file)

    def test_shred_paths_from_clipboard_menu_integration(self):
        """Shred a path copied to the real clipboard"""
        test_file = self.write_file('shred-me-via-real-clipboard')
        self.refresh_gui()

        options.set('expert_mode', True)
        options.set('delete_confirmation', False)

        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        clipboard.set_text(test_file, -1)
        self.refresh_gui()
        self.assertTrue(self.wait_until(
            lambda: test_file == clipboard.wait_for_text()))

        targets = [Gdk.atom_intern_static_string('text/plain')]
        self.app.cb_clipboard_uri_received(clipboard, targets, None)
        self.assertTrue(self.wait_until(lambda: not os.path.exists(test_file)))
        self.assertNotExists(test_file)

    def test_shred_clipboard_empty_no_glib_warnings(self):
        """Empty clipboard must not trigger GLib g_array warnings."""
        clear_clipboard()
        self.refresh_gui()

        with warnings.catch_warnings(record=True) as caught:
            warnings.simplefilter('always')
            self.app.cb_shred_clipboard(None, None)
            self.refresh_gui()
            glib_warnings = [
                warning for warning in caught
                if 'g_array' in str(warning.message)
            ]

        self.assertEqual([], glib_warnings)

    def _setup_new_cleaner(self, gui):
        def _create_cleaner_file_in_directory(dirname):
            cleaner_content = ('<?xml version="1.0" encoding="UTF-8"?>'
                               '<cleaner id="{}">'
                               '<label>Test run_operations</label>'
                               '<option id="{}">'
                               '<label>Test1</label>'
                               '<description>Delete files in a test directory</description>'
                               '<action command="delete" search="walk.all" path="{}"/>'
                               '</option>'
                               '</cleaner>'.format(self._NEW_CLEANER_ID, self._NEW_OPTION_ID, dirname))
            cleaner_filename = os.path.join(
                dirname, 'test_run_operations_cleaner.xml')
            self.write_file(cleaner_filename, cleaner_content, 'w')
            return cleaner_filename

        def _load_new_cleaner_in_gui(gui):
            # to load the new test cleaner with id 'test_run_operations'
            gui.cb_refresh_operations()
            self.refresh_gui()

        dirname = self.mkdtemp(prefix='bleachbit-test-run_operations')
        cleaner_filename = _create_cleaner_file_in_directory(dirname)
        self.assertExists(cleaner_filename)
        with mock.patch('bleachbit.CleanerML.list_cleanerml_files') as mock_list_cleanerml_files:
            with mock.patch('bleachbit.RecognizeCleanerML.cleaner_change_dialog') as mock_cleaner_change_dialog:
                mock_list_cleanerml_files.return_value = [cleaner_filename]
                mock_cleaner_change_dialog.return_value = None
                _load_new_cleaner_in_gui(gui)
        file_to_clean = self.mkstemp(prefix="somefile", dir=dirname)
        self.assertExists(file_to_clean)
        return file_to_clean

    def test_run_operations(self):
        gui = self.get_window()
        file_to_clean = self._setup_new_cleaner(gui)
        self._put_checkmark_on_cleaner(
            gui, self._NEW_CLEANER_ID, self._NEW_OPTION_ID)

        with mock.patch('bleachbit.GuiWindow.GUI._confirm_delete', return_value=True):
            self.assertTrue(gui._confirm_delete(False, False))
            # same as b = self.click_button(gui, _("Clean"))
            gui.run_operations(None)

        self.refresh_gui()
        self.assertNotExists(file_to_clean)

    def test_cb_run_option(self):
        gui = self.get_window()
        file_to_clean = self._setup_new_cleaner(gui)

        for really_delete, assert_method in [(False, self.assertExists), (True, self.assertNotExists)]:
            with mock.patch('bleachbit.GuiWindow.GUI._confirm_delete', return_value=True):
                self.assertTrue(gui._confirm_delete(False, False))
                gui.cb_run_option(
                    None, really_delete, self._NEW_CLEANER_ID, self._NEW_OPTION_ID
                )  # activated from context menu

            self.refresh_gui()
            assert_method(file_to_clean)

    def test_context_menu_cookie_manager(self):
        """Test cookie manager helper returns correct value for options"""
        gui = self.get_window()

        class _DummyAction:
            def __init__(self, action_key):
                self.action_key = action_key

        cookie_cleaner_id = '_unit_test_cookie_cleaner'
        non_cookie_cleaner_id = '_unit_test_non_cookie_cleaner'
        saved_cookie_cleaner = backends.get(cookie_cleaner_id)
        saved_non_cookie_cleaner = backends.get(non_cookie_cleaner_id)

        try:
            cookie_cleaner = type('DummyCleaner', (), {})()
            cookie_cleaner.actions = [('cookies', _DummyAction('cookie'))]
            non_cookie_cleaner = type('DummyCleaner', (), {})()
            non_cookie_cleaner.actions = [('logs', _DummyAction('delete'))]

            backends[cookie_cleaner_id] = cookie_cleaner
            backends[non_cookie_cleaner_id] = non_cookie_cleaner

            self.assertTrue(
                gui._option_has_cookie_command(cookie_cleaner_id, 'cookies'),
                'Expected test cookie cleaner option to be detected as cookie action')
            self.assertFalse(
                gui._option_has_cookie_command(non_cookie_cleaner_id, 'logs'),
                'Expected test non-cookie cleaner option to not be detected as cookie action')
        finally:
            if saved_cookie_cleaner is None:
                backends.pop(cookie_cleaner_id, None)
            else:
                backends[cookie_cleaner_id] = saved_cookie_cleaner
            if saved_non_cookie_cleaner is None:
                backends.pop(non_cookie_cleaner_id, None)
            else:
                backends[non_cookie_cleaner_id] = saved_non_cookie_cleaner

    def test_parent_unchecked_when_all_children_blocked_by_expert_mode(self):
        """When all children are blocked by expert mode, parent must stay unchecked."""

        model = Gtk.TreeStore(
            GObject.TYPE_STRING,   # 0: name
            GObject.TYPE_BOOLEAN,  # 1: active
            GObject.TYPE_PYOBJECT,  # 2: id
            GObject.TYPE_STRING,   # 3: size
            GObject.TYPE_STRING,   # 4: icon
        )
        parent_iter = model.append(
            None, ("TestCleaner", False, "test_cleaner", "", ""))
        model.append(parent_iter, ("Option1", False, "option1", "", ""))
        model.append(parent_iter, ("Option2", False, "option2", "", ""))

        tdm = TreeDisplayModel()

        def blocked_warning(_option_id):
            return "This option requires expert mode."

        with mock.patch('bleachbit.GuiTreeModels.backends',
                        {"test_cleaner": mock.Mock(get_warning=blocked_warning)}), \
                mock.patch('bleachbit.GuiTreeModels.options') as mock_options:
            mock_options.get.return_value = False  # expert_mode off
            mock_options.get_warning_preference.return_value = False

            parent_path = model.get_path(parent_iter)
            tdm.col1_toggled_cb(None, str(parent_path), model, None)

        self.assertFalse(model[parent_iter][1],
                         "Parent should remain unchecked when all children are blocked by expert mode")



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.


"""
Test case for module DeepScan
"""

# standard imports
import os
import shutil
import unittest
from unittest import mock

# first party imports
from tests import common
from tests import TestCleaner
from tests.common import SPECIAL_TEST_STRINGS
from bleachbit import IS_MAC, IS_POSIX, IS_WINDOWS
from bleachbit.Options import options
from bleachbit.DeepScan import DeepScan, Search, normalized_walk


class DeepScanTestCase(common.BleachbitTestCase):
    """Test Case for module DeepScan"""

    def _test_encoding(self, fn):
        """Test encoding"""

        fullpath = self.write_file(fn + ".bak")

        # Add Unicode paths to encourage a crash.
        subdir = os.path.join(self.tempdir, 'ɡælɪk.dir')
        os.mkdir(subdir)
        subfile = os.path.join(subdir, 'ɡælɪk.file')
        common.touch_file(subfile)

        searches = {self.tempdir: []}
        searches[self.tempdir].append(
            Search(command='delete', regex=r'\.[Bb][Aa][Kk]$'))
        ds = DeepScan(searches)
        found = False
        for cmd in ds.scan():
            if cmd == True:
                # True is used to yield to GTK+, but it is not
                # needed in this test.
                continue
            self.assertExists(cmd.path)
            if cmd.path == fullpath:
                found = True
        self.assertTrue(found, "Did not find '%s'" % fullpath)

        os.unlink(fullpath)
        self.assertNotExists(fullpath)

        shutil.rmtree(subdir)

    def test_encoding(self):
        """Test encoding"""
        for test in SPECIAL_TEST_STRINGS:
            self._test_encoding(test)

    def test_DeepScan(self):
        """Unit test for class DeepScan.  Preview real files."""
        path = os.path.expanduser('~')
        searches = {path: []}
        for regex in ('^Makefile$', '~$', 'bak$', '^Thumbs.db$', '^Thumbs.db:encryptable$'):
            searches[path].append(Search(command='delete', regex=regex))
        ds = DeepScan(searches)
        for cmd in ds.scan():
            if True == cmd:
                # it's yielding control to the GTK idle loop
                continue
            self.assertLExists(cmd.path)

    def test_skip_whitelisted_directory(self):
        """DeepScan should not search whitelisted directories."""
        # Reminder: each test case gets a new options directory,
        # so we do not need to restore the old keep list (whitelist).
        keep_dir = self.mkdir('keep')
        search_dir = self.mkdir('search')
        keep_file = self.write_file(os.path.join(keep_dir, 'skip.bbtestbak'))
        search_file = self.write_file(os.path.join(search_dir, 'find.bbtestbak'))

        options.set_whitelist_paths([('folder', keep_dir)])
        searches = {
            self.tempdir: [
                Search(command='delete', regex=r'\.bbtestbak$')
            ]
        }
        paths = [
            cmd.path
            for cmd in DeepScan(searches).scan()
            if cmd is not True
        ]
        self.assertNotIn(keep_file, paths)
        self.assertIn(search_file, paths)

    def _get_compiled_search(self, wholeregex_substr):
        """Return the CompiledSearch for the deepscan action whose
        wholeregex contains the given substring (e.g. 'node_modules',
        'venv')."""
        import xml.etree.ElementTree as ET
        from bleachbit.DeepScan import CompiledSearch

        xml_path = os.path.join(
            os.path.dirname(__file__), '..', 'cleaners', 'deepscan.xml')
        root = ET.parse(xml_path).getroot()
        wholeregex = None
        nwholeregex = None
        for action in root.iter('action'):
            if action.get('wholeregex') and wholeregex_substr in action.get('wholeregex'):
                wholeregex = action.get('wholeregex')
                nwholeregex = action.get('nwholeregex')
                break
        self.assertIsNotNone(wholeregex)
        self.assertIsNotNone(nwholeregex)
        return CompiledSearch(Search(command='delete', wholeregex=wholeregex, nwholeregex=nwholeregex))

    def test_node_modules_excludes_dot_dirs(self):
        """node_modules option should exclude system/cache/editor directories."""
        cs = self._get_compiled_search('node_modules')
        excluded = [
            ('/home/neo/.nvm/versions/node/v22.22.3/lib/node_modules/corepack', 'README.md'),
            ('/home/trinity/.vscode/extensions/ms-python.python-2026.4.0-linux-x64/out/client/node_modules', 'sudo-prompt.js'),
            ('/home/morpheus/.npm/_npx/15b07286cbcc3329/node_modules',
             '.package-lock.json'),
            ('/home/smith/.cache/typescript/5.9/node_modules', '.package-lock.json'),
            (r'C:\Users\glados\.windsurf\extensions\ms-python.python-2026.4.0-universal\out\client\node_modules', 'unicode.js'),
            (r'C:\Users\glados\.cursor\projects\empty-window\canvases\node_modules', 'cursor.js'),
            (r'C:\Users\glados\AppData\Local\Programs\myapp\node_modules', 'file.js'),
            (r'C:\Users\glados\AppData\Local\ApertureScience\MainframeCore\node_modules\central-processing', 'index.js'),
            (r'C:\Users\Alice\AppData\Local\Programs\BeeperTexts\resources\app.asar.unpacked\node_modules\@beeper',
             'beeper_client_sdk_napi_binding.node'),
            (r'C:\Users\bob\AppData\Local\Programs\Microsoft VS Code\7e7950df89\resources\app',
             'node_modules.asar'),
            (r'C:\Users\mallory\AppData\Local\Programs\Microsoft VS Code\fcf604774b\resources\app',
             'node_modules.asar'),
            ('/home/trent/projects/important/node_modules_backup', 'README.md'),
            ('/home/trent/projects/important', 'node_modules.asar'),
            (r'C:\Users\charlie\Documents\important', 'node_modules.asar'),
            (r'C:\Users\charlie\Documents\important\node_modules_backup', 'aes.js'),
        ]
        for dirpath, filename in excluded:
            with self.subTest(dirpath=dirpath, filename=filename):
                self.assertIsNone(
                    cs.match(dirpath, filename),
                    f"Should exclude {dirpath}/{filename}")

    def test_node_modules_matches_ordinary_dirs(self):
        """node_modules option should match ordinary user project directories."""
        cs = self._get_compiled_search('node_modules')
        matched = [
            ('/home/sfalken/games/wopr/node_modules/joshua/ai', 'core.js'),
            (r'C:\Users\esnowden\Documents\nsa\prism\node_modules\crypto-js', 'aes.js'),
            (r'C:\Users\chunkylover53\Documents\work-from-home\node_modules\click-button-automator\node_modules\is-buffer', 'index.js'),
            ('/home/chunkylover53/Downloads/donut-radar/node_modules/react-dom', 'bar.js'),
        ]
        for dirpath, filename in matched:
            with self.subTest(dirpath=dirpath, filename=filename):
                self.assertIsNotNone(
                    cs.match(dirpath, filename),
                    f"Should match {dirpath}/{filename}")

    def test_venv_excludes_system_and_editor_dirs(self):
        """venv option should exclude system/cache/editor directories and non-venv paths."""
        cs = self._get_compiled_search('venv')
        excluded = [
            # venv cannot be directly under home
            ('/home/username/.venv', 'pyvenv.cfg'),
            # Python standard library venv module (not a virtual environment)
            ('/home/username/.local/share/uv/python/cpython-3.12.11-linux-x86_64-gnu/lib/python3.12/venv',
             '__init__.py'),
            ('/home/username/.pyenv/versions/3.12.7/lib/python3.12/venv', '__init__.py'),
            # pipx-managed venvs (managed by pipx, like npm cache)
            ('/home/username/.local/share/pipx/venvs/duplicity', 'pyvenv.cfg'),
            # typeshed stubs bundled with editor extensions
            ('/home/username/.vscode/extensions/ms-python.python-2026.4.0-linux-x64/python_files/lib/jedilsp/jedi/third_party/typeshed/stdlib/3/venv',
             '__init__.pyi'),
            # pyenv build patches
            ('/home/username/.pyenv/plugins/python-build/share/python-build/patches/3.5.2/Python-3.5.2',
             'venv.patch'),
            # Windows: editor extensions and AppData
            (r'C:\Users\glados\.vscode\extensions\ms-python.python-2026.4.0-universal\python_files\lib\jedilsp\jedi\third_party\typeshed\stdlib\3\venv',
             '__init__.pyi'),
            (r'C:\Users\glados\AppData\Local\ApertureScience\MainframeCore\venv', 'pyvenv.cfg'),
            # Files/directories that contain venv (but not a venv directory)
            ('/home/aturing/npl/ace_project/tools/', 'bootstrap_ace_venv.sh'),
            ('/home/vonneumann/edvac/logs/', 'venv_build.log'),
            ('/home/aturing/npl/ace_project/venv_tools/',
             'requirements_numerics.txt'),
            (r'C:\Users\charlie\Documents\important\venv_backup', 'pyvenv.cfg'),
            # .venv bundled inside an AppData\Local application directory
            (r'C:\Users\charlie\AppData\Local\AnkiProgramFiles\.venv\Lib\site-packages\PyQt6',
             'QtCore.pyd'),
            # Python standard library venv module under uv-managed Python (Windows)
            (r'C:\Users\charlie\AppData\Roaming\uv\python\cpython-3.13.5-windows-x86_64-none\Lib\venv',
             '__main__.py'),
        ]
        if IS_WINDOWS:
            # Windows is case insensitive
            # venv cannot be directly under home
            excluded.append((r'c:\users\username\.VENV', 'pyvenv.cfg'))
        for dirpath, filename in excluded:
            with self.subTest(dirpath=dirpath, filename=filename):
                self.assertIsNone(
                    cs.match(dirpath, filename),
                    f"Should exclude {dirpath}/{filename}")

    def test_venv_matches_ordinary_dirs(self):
        """venv option should match ordinary user project virtual environments."""
        cs = self._get_compiled_search('venv')
        matched = [
            ('/home/genisys/t800/.venv/yolo-v666', 'pyvenv.cfg'),
            ('/home/stark/jarvis/venv/yolo14', 'pyvenv.cfg'),
            (r'C:\Users\esnowden\Documents\nsa\prism\venv', 'pyvenv.cfg'),
        ]
        if IS_WINDOWS:
            # case insensitive
            matched.append(
                (r'C:\USERS\esnowden\Documents\nsa\prism\.VENV', 'pyvenv.cfg'))
            matched.append(
                (r'c:\USERS\pinky\world_domination\VENV\site-packages\hypnosisis', '__init__.py'))
        for dirpath, filename in matched:
            with self.subTest(dirpath=dirpath, filename=filename):
                self.assertIsNotNone(
                    cs.match(dirpath, filename),
                    f"Should match {dirpath}/{filename}")

    def _test_delete(self, command):
        """Delete files in a test environment"""

        # make some files
        f_del = self.write_file('foo.txt.bbtestbak')
        f_keep = self.write_file('foo.txt')
        subdir = os.path.join(self.tempdir, 'sub')
        os.mkdir(subdir)
        f_del2 = self.write_file(os.path.join(subdir, 'bar.ini.bbtestbak'))
        f_del3 = self.write_file(os.path.join(subdir, 'bar.ini.bbtestBAK'))

        # sanity check
        self.assertExists(f_del)
        self.assertExists(f_keep)
        self.assertExists(f_del2)

        # run deep scan
        def run_deep_scan(regex):
            astr = '<action command="{}" search="deep" cache="false" path="{}" {}/>'.format(
                command, self.tempdir, regex)
            cleaner = TestCleaner.action_to_cleaner(astr)
            from bleachbit.Worker import backends, Worker
            backends['test'] = cleaner
            operations = {'test': ['option1']}
            from bleachbit import CLI
            ui = CLI.CliCallback()
            worker = Worker(ui, True, operations)
            list(worker.run())
            del backends['test']

        # validate results
        run_deep_scan(r'regex="\.bbtestbak$" wholeregex="sub.*\.bbtestbak$"')
        self.assertExists(f_del)
        self.assertExists(f_keep)
        self.assertFalse(os.path.exists(f_del2))
        if IS_POSIX:
            self.assertExists(f_del3)
        else:
            self.assertFalse(os.path.exists(f_del3))

        # validate results
        run_deep_scan(r'regex="\.bbtestbak$"')
        self.assertFalse(os.path.exists(f_del))
        self.assertExists(f_keep)

        # cleanup
        os.unlink(f_keep)
        if IS_POSIX:
            os.unlink(f_del3)
        os.rmdir(subdir)

    def test_delete(self):
        self._test_delete('delete')

    def test_shred(self):
        self._test_delete('shred')

    @unittest.skipUnless(IS_MAC, 'Not on Darwin')
    def test_normalized_walk_darwin(self):

        with mock.patch('os.walk') as mock_walk:
            mock_walk.return_value = [
                ('/foo', ('bar',), ['ba\xcc\x80z']),
                ('/foo/bar', (), ['spam', 'eggs']),
            ]
            with mock.patch('platform.system') as mock_platform_system:
                mock_platform_system.return_value = 'Darwin'
                self.assertEqual(list(normalized_walk('.')), [
                    ('/foo', ('bar',), ['b\xc3\xa0z']),
                    ('/foo/bar', (), ['spam', 'eggs']),
                ])

        with mock.patch('os.walk') as mock_walk:
            expected = [
                ('/foo', ('bar',), ['baz']),
                ('/foo/bar', (), ['spam', 'eggs']),
            ]
            mock_walk.return_value = expected
            self.assertEqual(list(normalized_walk('.')), expected)



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.

"""
Launcher
"""

import os
import sys


def _add_posix_share_to_path():
    """Setup path on POSIX systems for AppImage"""
    if os.name != 'posix':
        return
    # Add the appropriate share directory to sys.path so we can import bleachbit.
    # This must happen before importing from bleachbit.Bootstrap.
    #
    _launcher_dir = os.path.dirname(os.path.abspath(__file__))
    if not (os.path.isfile(os.path.join(_launcher_dir, 'bleachbit', '__init__.py')) and
            os.path.isfile(os.path.join(_launcher_dir, 'bleachbit', 'Unix.py'))):
        _appdir = os.environ.get('APPDIR')
        if _appdir:
            _share = os.path.normpath(os.path.join(_appdir, 'usr/share/'))
            if os.path.isfile(os.path.join(_share, 'bleachbit', '__init__.py')):
                sys.path.insert(0, _share)
        elif os.path.isfile('/usr/share/bleachbit/__init__.py'):
            sys.path.append('/usr/share/')


def main():
    """Main entry point"""
    _add_posix_share_to_path()
    from bleachbit.Bootstrap import bootstrap, check_wayland_and_root
    bootstrap()

    # Check for GUI only when needed: this avoids a Gtk warning when
    # a display is not available.
    if len(sys.argv) == 1 or '--gui' in sys.argv:
        from bleachbit.GtkShim import HAVE_GTK
        have_gui = HAVE_GTK
    else:
        have_gui = False

    # If no arguments provided and display is available, then
    # assume the user wants to start the GUI.
    # If Wayland is a problem and no explicit CLI args were provided,
    # then fall back to CLI.
    if 1 == len(sys.argv) and have_gui and not check_wayland_and_root():
        # Import GUI inside the condition for Linux packagers to
        # separate GUI into another package.
        import bleachbit.GuiApplication  # pylint: disable=import-outside-toplevel
        app = bleachbit.GuiApplication.Bleachbit()
        sys.exit(app.run(sys.argv))

    # Either CLI args were provided or no display is available
    import bleachbit.CLI  # pylint: disable=import-outside-toplevel
    # If no args and defaulting to CLI, print usage information
    if 1 == len(sys.argv) and not have_gui:
        sys.argv.append('--help')
    try:
        bleachbit.CLI.process_cmd_line()
    except BrokenPipeError:
        # The downstream pipe consumer (e.g., `less`) closed before we
        # finished writing.  Redirect stdout to devnull to suppress the
        # secondary BrokenPipeError raised by the interpreter while
        # flushing stdout at shutdown, then exit with the conventional
        # SIGPIPE status (128 + 13).
        devnull = os.open(os.devnull, os.O_WRONLY)
        os.dup2(devnull, sys.stdout.fileno())
        sys.exit(141)


if __name__ == '__main__':
    main()



# vim: ts=4:sw=4:expandtab

# BleachBit
# Copyright (C) 2008-2025 Andrew Ziem
# https://www.bleachbit.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Test case for Command
"""

import os
from unittest import mock

from tests import common
from bleachbit import FileUtilities
from bleachbit.Command import Delete, Function, Shred


class CommandTestCase(common.BleachbitTestCase):
    """Test case for Command"""

    def test_Delete(self, cls=Delete):
        """Unit test for Delete"""
        path = self.write_file('test_Delete', b'foo')
        cmd = cls(path)
        self.assertExists(path)

        # preview
        ret = next(cmd.execute(really_delete=False))
        self.assertGreater(ret['size'], 0)
        self.assertEqual(ret['path'], path)
        self.assertExists(path)

        # delete
        ret = next(cmd.execute(really_delete=True))
        self.assertGreater(ret['size'], 0)
        self.assertEqual(ret['path'], path)
        self.assertNotExists(path)

    def test_delete_permission(self):
        """Test delete with permission denied for getsize()"""
        path = self.write_file('test_delete_permission', b'foo')
        cmd = Delete(path)
        self.assertExists(path)

        with mock.patch('bleachbit.FileUtilities.getsize') as mock_getsize:
            mock_getsize.side_effect = PermissionError('Permission denied')

            # preview
            ret = next(cmd.execute(really_delete=False))
            self.assertIsNone(ret['size'])
            self.assertEqual(ret['path'], path)
            self.assertExists(path)

    def test_Function(self):
        """Unit test for Function"""
        path = self.write_file('test_Function', b'foo')
        cmd = Function(path, FileUtilities.delete, 'bar')
        self.assertExists(path)
        self.assertGreater(os.path.getsize(path), 0)

        # preview
        ret = next(cmd.execute(False))
        self.assertExists(path)
        self.assertGreater(os.path.getsize(path), 0)

        # delete
        ret = next(cmd.execute(True))
        self.assertGreater(ret['size'], 0)
        self.assertEqual(ret['path'], path)
        self.assertNotExists(path)

    def test_Function_no_collation(self):
        """Unit test for Function with no collation

        See https://github.com/bleachbit/bleachbit/issues/1866
        """
        path = self.write_file('test_Function_no_collation', b'')
        cmd = Function(path,
                       lambda p: FileUtilities.execute_sqlite3(
                           p, 'CREATE TABLE test (name TEXT COLLATE foo);'),
                       'test_no_collation')

        with mock.patch('bleachbit.Command.logger.debug') as mock_debug:
            with self.assertRaises(StopIteration):
                next(cmd.execute(True))
            mock_debug.assert_called_with(mock.ANY)

    def test_Shred(self):
        """Unit test for Shred"""
        self.test_Delete(Shred)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.

"""
Test case for module Windows
"""


# standard imports
import ctypes
import itertools
import os
import platform
import shutil
import subprocess
import sys
import tempfile
import time
from decimal import Decimal
from pathlib import Path
from unittest import mock
from random import randint

# first party imports
from tests import common

import bleachbit
from bleachbit import FileUtilities, General
from bleachbit.Command import Delete, Function
from bleachbit.FileUtilities import extended_path, extended_path_undo
from bleachbit.Windows import (
    delete_locked_file,
    delete_registry_key,
    delete_registry_value,
    delete_updates,
    is_service_running,
    run_net_service_command,
    detect_registry_key,
    empty_recycle_bin,
    get_clipboard_paths,
    get_fixed_drives,
    get_font_conf_file,
    get_known_folder_path,
    get_recycle_bin,
    get_windows_system_paths,
    get_windows_version,
    is_junction,
    move_to_recycle_bin,
    parse_windows_build,
    path_on_network,
    set_environ,
    setup_environment,
    shell_change_notify,
    split_registry_key,
    read_registry_key,
    get_sid_token_48,
    is_ots_elevation,
    get_splash_screen_delay_seconds,
    expand_windows_system_vars,
    SplashThread,
)
from bleachbit import logger


class WindowsSystemPathsTestCase(common.BleachbitTestCase):
    """Test Windows system directory expansion."""

    def test_get_windows_system_paths_32_bit_os(self):
        """Unit test get_windows_system_paths() for 32-bit Windows."""
        env = {
            'WinDir': r'C:\Windows',
            'PROCESSOR_ARCHITECTURE': 'x86',
        }
        self.assertEqual(
            [r'C:\Windows\System32'],
            get_windows_system_paths(env, process_bits=32))

    def test_get_windows_system_paths_64_bit_process(self):
        """Unit test get_windows_system_paths() for 64-bit Python."""
        env = {
            'WinDir': r'C:\Windows',
            'PROCESSOR_ARCHITECTURE': 'AMD64',
        }
        self.assertEqual(
            [r'C:\Windows\System32', r'C:\Windows\SysWOW64'],
            get_windows_system_paths(env, process_bits=64))

    def test_get_windows_system_paths_wow64_process(self):
        """Unit test get_windows_system_paths() for 32-bit Python on 64-bit Windows."""
        env = {
            'WinDir': r'C:\Windows',
            'PROCESSOR_ARCHITECTURE': 'x86',
            'PROCESSOR_ARCHITEW6432': 'AMD64',
        }
        self.assertEqual(
            [r'C:\Windows\Sysnative', r'C:\Windows\SysWOW64'],
            get_windows_system_paths(env, process_bits=32))

    def test_get_windows_system_paths(self):
        """Test get_windows_system_paths() in general"""
        # Test with default environment
        ret = get_windows_system_paths()
        self.assertIsInstance(ret, list)
        for path in ret:
            self.assertTrue(os.path.isabs(path))
            self.assertExists(path)

    def test_expand_windows_system_vars(self):
        """Unit test expand_windows_system_vars()."""
        path_arg = (r'%WindowsSystem%\LogFiles\*.log')
        # Without providing system paths
        paths = expand_windows_system_vars(path_arg)
        self.assertIsInstance(paths, list)
        # On 32-bit OS, it returns one. On 64-bit OS, it returns
        # 2 (regardless of bitness of the process).
        self.assertIn(len(paths), (1, 2))
        # Provide system paths
        paths = expand_windows_system_vars(
            path_arg,
            [r'C:\Windows\Sysnative', r'C:\Windows\SysWOW64'])
        self.assertEqual(
            [r'C:\Windows\Sysnative\LogFiles\*.log',
             r'C:\Windows\SysWOW64\LogFiles\*.log'],
            paths)


if bleachbit.IS_WINDOWS:
    import pywintypes
    import win32api
    import win32service
    import winreg
    from win32com.shell import shell

    from bleachbit import Windows


def put_objects_into_recycle_bin():
    """Put a file and a folder into the recycle bin"""
    count = 0
    # make a file and move it to the recycle bin
    tests = ['regular', 'unicode-emdash-u\u2014',
             'long' + 'x' * 100] + common.SPECIAL_TEST_STRINGS
    for test in tests:
        with tempfile.NamedTemporaryFile(
                prefix='bleachbit-recycle-file', suffix=test, delete=False) as f:
            filename = f.name
            # Write enough data to exceed the MFT resident threshold.
            f.write(b'x' * (1024 * 32))
        move_to_recycle_bin(filename)
        count += 1
    # make a folder and move it to the recycle bin
    dirname = tempfile.mkdtemp(prefix='bleachbit-recycle-folder')
    common.touch_file(os.path.join(dirname, 'file'))
    move_to_recycle_bin(dirname)
    count += 1
    return count


class WindowsLinksMixIn():
    """Mixin class for Windows link creation methods."""

    def _create_win_dir_symlink(self, target, linkname):
        """Create a directory symlink"""

        self.assertFalse(os.path.lexists(linkname),
                         f'Link must not exist: {linkname}')
        self.assertTrue(os.path.isabs(target),
                        f'Target must be absolute path: {target}')
        self.assertTrue(os.path.isabs(linkname),
                        f'Link must be absolute path: {linkname}')
        self.assertExists(target)
        target_path = Path(target)
        self.assertTrue(target_path.is_dir(),
                        f'Target must be an existing directory: {target}')

        kernel32 = ctypes.windll.kernel32
        kernel32.CreateSymbolicLinkW.argtypes = [
            ctypes.c_wchar_p,
            ctypes.c_wchar_p,
            ctypes.c_uint32,
        ]
        kernel32.CreateSymbolicLinkW.restype = ctypes.c_ubyte
        result = kernel32.CreateSymbolicLinkW(
            linkname, target, 1)  # SYMBOLIC_LINK_FLAG_DIRECTORY
        if result == 0:
            err = ctypes.GetLastError()
            raise OSError(err, ctypes.FormatError(err))
        self.assertExists(linkname)
        link_path = Path(linkname)
        self.assertTrue(link_path.is_symlink())
        self.assertTrue(link_path.is_dir())
        self.assertFalse(Windows.is_junction(linkname))
        self.assertFalse(FileUtilities.is_normal_directory(linkname))

    def _create_win_hard_link(self, target, linkname):
        """Create a hard link to a file"""

        self.assertFalse(os.path.lexists(linkname),
                         f'Link must not exist: {linkname}')
        self.assertTrue(os.path.isabs(target),
                        f'Target must be absolute path: {target}')
        self.assertTrue(os.path.isabs(linkname),
                        f'Link must be absolute path: {linkname}')
        self.assertExists(target)
        target_path = Path(target)
        self.assertTrue(target_path.is_file(),
                        f'Target must be an existing file: {target}')
        kernel32 = ctypes.windll.kernel32
        kernel32.CreateHardLinkW.argtypes = [
            ctypes.c_wchar_p,
            ctypes.c_wchar_p,
            ctypes.c_void_p,
        ]
        kernel32.CreateHardLinkW.restype = ctypes.c_bool
        result = kernel32.CreateHardLinkW(linkname, target, None)
        if not result:
            err = ctypes.GetLastError()
            raise OSError(err, ctypes.FormatError(err))
        self.assertExists(linkname)
        link_path = Path(linkname)
        self.assertTrue(link_path.is_file())
        self.assertFalse(link_path.is_symlink())
        self.assertFalse(Windows.is_junction(linkname))
        self.assertTrue(os.path.samefile(target, linkname))
        self.assertFalse(FileUtilities.is_normal_directory(linkname))

    def _create_win_junction(self, target, linkname):
        """Create a directory junction using mklink /J."""

        if os.path.lexists(linkname):
            raise OSError(f'Link already exists: {linkname}')
        self.assertTrue(os.path.isabs(target),
                        f'Target must be absolute path: {target}')
        self.assertTrue(os.path.isabs(linkname),
                        f'Link must be absolute path: {linkname}')
        self.assertExists(target)
        target_path = Path(target)
        self.assertTrue(target_path.is_dir(),
                        f'Target must be an existing directory: {target}')
        cmd = ['cmd', '/c', 'mklink', '/J',
               extended_path(linkname), extended_path(target)]
        subprocess.check_call(cmd)
        self.assertExists(linkname)
        self.assertTrue(Windows.is_junction(linkname))
        path = Path(linkname)
        self.assertTrue(path.is_dir())
        self.assertFalse(path.is_symlink())


@common.skipUnlessWindows
class WindowsTestCase(common.BleachbitTestCase, WindowsLinksMixIn):

    """Test case for module Windows"""

    def skipUnlessAdmin(self):
        if not shell.IsUserAnAdmin():
            self.skipTest('requires administrator privileges')

    def test_get_recycle_bin(self):
        """Unit test for get_recycle_bin"""
        for f in get_recycle_bin():
            self.assertLExists(extended_path(f))

    @common.skipUnlessDestructive
    def test_get_recycle_bin_destructive(self):
        """Unit test the destructive part of get_recycle_bin"""
        added_count = put_objects_into_recycle_bin()
        # clear recycle bin
        counter = 0
        for f in get_recycle_bin():
            counter += 1
            FileUtilities.delete(f)
        self.assertGreaterEqual(counter, added_count, 'deleted %d' % counter)
        # now it should be empty
        for _f in get_recycle_bin():
            self.fail('recycle bin should be empty, but it is not')

    def _test_link_helper(self, mklink_option, recycle_container, clear_recycle_bin):
        """Helper function for testing directory junctions and symlinks in the recycle bin.

        Tests that get_recycle_bin() correctly handles Windows directory links
        when they are placed in the recycle bin, and verifies that the original
        target directory and its contents are preserved after clearing the bin.

        Args:
            mklink_option: Link type to create. '/j' for directory junction,
                '/d' for directory symbolic link.
            recycle_container: If True, move the container directory to the
                recycle bin. If False, move only the link itself.
            clear_recycle_bin: If True, clear the recycle bin after moving
                the link. If False, manually delete items from get_recycle_bin().

        Note:
            Directory symbolic links (/d) require administrator privileges.
        """
        assert mklink_option in ('/j', '/d')
        if mklink_option == '/d':
            self.skipUnlessAdmin()
        # make a normal directory with a file in it
        target_dir = self.mkdir('target_dir')

        canary_base = f'do_not_delete{randint(10000, 9999999)}'
        canary_fn = os.path.join(target_dir, canary_base)
        common.touch_file(canary_fn)
        self.assertFalse(is_junction(canary_fn))

        # make a normal directory to hold a link
        container_dir = self.mkdir('container_dir')

        # create the link
        link_pathname = os.path.join(container_dir, 'link')
        if mklink_option == '/j':
            self._create_win_junction(target_dir, link_pathname)
        else:
            self._create_win_dir_symlink(target_dir, link_pathname)

        # put the link in the recycle bin
        move_to_recycle_bin(
            container_dir if recycle_container else link_pathname)

        def cleanup_dirs():
            shutil.rmtree(container_dir, True)
            self.assertNotExists(container_dir)
            shutil.rmtree(target_dir, True)

        if not clear_recycle_bin:
            cleanup_dirs()
            return

        # clear the recycle bin
        for f in get_recycle_bin():
            if canary_base in f:
                logger.error('get_recycle_bin() returned canary: %s', f)
            FileUtilities.delete(f, shred=False)

        # verify the canary is still there
        self.assertExists(canary_fn)

        # clean up
        cleanup_dirs()

    def test_link_types(self):
        """Unit test for directory junctions and symlinks with recycle bin"""
        for mklink_option, recycle_container, clear_recycle_bin in itertools.product(
            ('/j', '/d'), (False, True), (False, True)
        ):
            with self.subTest(mklink_option=mklink_option,
                              recycle_container=recycle_container,
                              clear_recycle_bin=clear_recycle_bin):
                self._test_link_helper(
                    mklink_option, recycle_container, clear_recycle_bin)

    def _test_broken_link_in_recycle_bin(self, mklink_option, recycle_container,
                                         break_before_recycle):
        """Helpful function to test broken directory junctions and symlinks
        in the recycle bin.

        When a broken link (target deleted) is in the recycle bin, deleting it
        should delete the link itself, not attempt to follow or restore the
        target.

        Args:
            mklink_option: Link type to create. '/j' for directory junction,
                '/d' for directory symbolic link.
            recycle_container: If True, move the container directory to the
                recycle bin. If False, move only the link itself.
            break_before_recycle: If True, delete the target before moving
                the link to the recycle bin. If False, recycle first and then
                delete the target.
        """
        assert mklink_option in ('/j', '/d')
        if mklink_option == '/d':
            self.skipUnlessAdmin()

        before = set(get_recycle_bin())
        target_dir = self.mkdir('target_dir')
        canary_fn = os.path.join(
            target_dir, f'canary{randint(10000, 9999999)}')
        common.touch_file(canary_fn)

        container_dir = self.mkdir('container_dir')
        link_pathname = os.path.join(container_dir, 'link')
        if mklink_option == '/j':
            self._create_win_junction(target_dir, link_pathname)
        else:
            self._create_win_dir_symlink(target_dir, link_pathname)

        self.assertTrue(os.path.lexists(link_pathname))
        self.assertExists(canary_fn)

        if break_before_recycle:
            shutil.rmtree(target_dir)
            self.assertTrue(os.path.lexists(link_pathname))
            self.assertFalse(os.path.exists(link_pathname))

        move_to_recycle_bin(
            container_dir if recycle_container else link_pathname)

        if not break_before_recycle:
            shutil.rmtree(target_dir)
            if not recycle_container:
                shutil.rmtree(container_dir)

        try:
            added = [p for p in get_recycle_bin() if p not in before]
            self.assertTrue(
                added, 'recycle bin should contain recycled link')

            broken = []
            for f in added:
                ep = extended_path(f)
                self.assertLExists(ep)
                if ((os.path.islink(ep) or is_junction(ep)) and
                        not os.path.exists(ep)):
                    broken.append(f)
            self.assertTrue(
                broken, 'expected broken link (lexists, not exists)')

            for f in added:
                ep = extended_path(f)
                self.assertTrue(FileUtilities.delete(ep, shred=False))
                self.assertNotLExists(ep)

            self.assertFalse(os.path.lexists(link_pathname))
            self.assertFalse(os.path.lexists(target_dir))
            if recycle_container or not break_before_recycle:
                self.assertFalse(os.path.lexists(container_dir))
            else:
                self.assertTrue(os.path.lexists(container_dir))
        finally:
            empty_recycle_bin(None, True)
            if os.path.lexists(container_dir):
                shutil.rmtree(container_dir, True)
            if os.path.lexists(target_dir):
                shutil.rmtree(target_dir, True)

    def _test_broken_link_delete(self, mklink_option, shred):
        """Helper function to test FileUtilities.delete() on a broken
        link outside the recycle bin.

        - Tests both junctions and directory symlinks.
        - Test is Windows only.
        - Target is a directory (not a file).
        - Target is broken (does not exist).
        - Tests both shredding and unlink.
        - Does not invoke the recycle bin.
        """
        assert mklink_option in ('/j', '/d')
        if mklink_option == '/d':
            self.skipUnlessAdmin()

        target_dir = self.mkdir('broken_link_delete_target')
        container_dir = self.mkdir('broken_link_delete_container')
        link_pathname = os.path.join(container_dir, 'link')
        if mklink_option == '/j':
            self._create_win_junction(target_dir, link_pathname)
        else:
            self._create_win_dir_symlink(target_dir, link_pathname)
        shutil.rmtree(target_dir)
        self.assertLExists(link_pathname)
        self.assertFalse(os.path.exists(link_pathname))
        self.assertTrue(FileUtilities.delete(link_pathname, shred=shred))
        self.assertNotLExists(link_pathname)
        shutil.rmtree(container_dir, True)

    def test_broken_link_in_recycle_bin(self):
        """Unit test for broken directory junctions and symlinks in recycle bin

        - Tests both junctions and directory symlinks.
        - Target is a directory (not a file).
        - Target is broken (does not exist).
        - Tests both container in recycle bin and outside recycle bin.
        - Tests both break target before and after recycling.
        - Test is Windows only (of course).
        """
        for mklink_option, recycle_container, break_before_recycle in itertools.product(
            ('/j', '/d'), (False, True), (False, True)
        ):
            with self.subTest(mklink_option=mklink_option,
                              recycle_container=recycle_container,
                              break_before_recycle=break_before_recycle):
                self._test_broken_link_in_recycle_bin(
                    mklink_option, recycle_container, break_before_recycle)
        for mklink_option, shred in itertools.product(
            ('/j', '/d'), (False, True)
        ):
            with self.subTest(mklink_option=mklink_option, shred=shred,
                              outside_bin=True):
                self._test_broken_link_delete(mklink_option, shred)

    def test_delete_locked_file(self):
        """Unit test for delete_locked_file"""
        tests = ('regular', 'unicode-emdash-u\u2014', 'long' + 'x' * 100)
        for test in tests:
            pathname = self.mkstemp(suffix=test)
            time.sleep(5)  # avoid race condition
            self.assertExists(pathname)
            logger.debug('delete_locked_file(%s) ' % pathname)
            if not shell.IsUserAnAdmin():
                with self.assertRaises(WindowsError):
                    delete_locked_file(pathname)
            else:
                try:
                    delete_locked_file(pathname)
                except WindowsError:
                    logger.exception(
                        'delete_locked_file() threw an error, which may be a false positive')
            self.assertExists(pathname)
        logger.info('reboot Windows and check the three files are deleted')

    def test_delete_parent_lock_needed(self):
        old_admin = Windows._delete_parent_lock_admin
        try:
            with mock.patch('bleachbit.Windows.shell.IsUserAnAdmin', return_value=False):
                Windows._delete_parent_lock_admin = None
                self.assertFalse(Windows._delete_parent_lock_needed(
                    r'C:\Windows\Temp\bleachbit-test-file'))

            with mock.patch('bleachbit.Windows.shell.IsUserAnAdmin', return_value=True):
                Windows._delete_parent_lock_admin = None
                with mock.patch.dict(os.environ, {'USERPROFILE': r'C:\Users\bleachbit'}):
                    self.assertFalse(Windows._delete_parent_lock_needed(
                        r'C:\Users\bleachbit\AppData\Local\Temp\bleachbit-test-file'))
                    self.assertTrue(Windows._delete_parent_lock_needed(
                        r'C:\Windows\Temp\bleachbit-test-file'))
        finally:
            Windows._delete_parent_lock_admin = old_admin

    def test_delete_with_parent_lock_reuses_handle(self):
        Windows._delete_parent_lock_handle = None
        Windows._delete_parent_lock_key = None
        handle = mock.sentinel.handle
        delete_func = mock.Mock(return_value=mock.sentinel.deleted)
        with mock.patch('bleachbit.Windows._delete_parent_lock_needed', return_value=True), \
                mock.patch('bleachbit.Windows._delete_parent_directory', return_value=r'\\?\C:\Windows\Temp'), \
                mock.patch('bleachbit.Windows.win32file.CreateFile', return_value=handle) as create_file, \
                mock.patch('bleachbit.Windows.win32file.GetFileAttributesW', return_value=0), \
                mock.patch('bleachbit.Windows.win32file.CloseHandle') as close_handle:
            self.assertEqual(Windows.with_parent_lock(
                r'C:\Windows\Temp\one.tmp', delete_func, r'C:\Windows\Temp\one.tmp'), mock.sentinel.deleted)
            self.assertEqual(Windows.with_parent_lock(
                r'C:\Windows\Temp\two.tmp', delete_func, r'C:\Windows\Temp\two.tmp'), mock.sentinel.deleted)
            self.assertEqual(create_file.call_count, 1)
            self.assertEqual(delete_func.call_count, 2)
            Windows._close_delete_parent_lock()
            close_handle.assert_called_once_with(handle)

    def test_delete_with_parent_lock_closes_handle_on_delete_error(self):
        Windows._delete_parent_lock_handle = None
        Windows._delete_parent_lock_key = None
        handle = mock.sentinel.handle
        delete_func = mock.Mock(side_effect=RuntimeError('delete failed'))
        with mock.patch('bleachbit.Windows._delete_parent_lock_needed', return_value=True), \
                mock.patch('bleachbit.Windows._delete_parent_directory', return_value=r'\\?\C:\Windows\Temp'), \
                mock.patch('bleachbit.Windows.win32file.CreateFile', return_value=handle), \
                mock.patch('bleachbit.Windows.win32file.GetFileAttributesW', return_value=0), \
                mock.patch('bleachbit.Windows.win32file.CloseHandle') as close_handle:
            with self.assertRaises(RuntimeError):
                Windows.with_parent_lock(
                    r'C:\Windows\Temp\one.tmp', delete_func, r'C:\Windows\Temp\one.tmp')
            close_handle.assert_called_once_with(handle)
            self.assertIsNone(Windows._delete_parent_lock_handle)

    def test_lock_delete_parent_rejects_reparse_point(self):
        Windows._delete_parent_lock_handle = None
        Windows._delete_parent_lock_key = None
        handle = mock.sentinel.handle
        with mock.patch('bleachbit.Windows._delete_parent_directory', return_value=r'\\?\C:\Windows\Temp'), \
                mock.patch('bleachbit.Windows.win32file.CreateFile', return_value=handle), \
                mock.patch('bleachbit.Windows.win32file.GetFileAttributesW',
                           return_value=Windows.FILE_ATTRIBUTE_REPARSE_POINT), \
                mock.patch('bleachbit.Windows.win32file.CloseHandle') as close_handle:
            with self.assertRaises(OSError):
                Windows._lock_delete_parent(r'C:\Windows\Temp\one.tmp')
            close_handle.assert_called_once_with(handle)
            self.assertIsNone(Windows._delete_parent_lock_handle)

    def test_delete_registry_key(self):
        """Unit test for delete_registry_key"""
        # (return value, key, really_delete)
        tests = ((False, 'HKCU\\Software\\BleachBit\\DoesNotExist', False, ),
                 (False, 'HKCU\\Software\\BleachBit\\DoesNotExist', True, ),
                 (True, 'HKCU\\Software\\BleachBit\\DeleteThisKey', False, ),
                 (True, 'HKCU\\Software\\BleachBit\\DeleteThisKey', True, ), )

        # create a nested key
        key = 'Software\\BleachBit\\DeleteThisKey'
        subkey = key + '\\AndThisKey'
        hkey = winreg.CreateKey(winreg.HKEY_CURRENT_USER, subkey)
        hkey.Close()

        # test
        for test in tests:
            rc = test[0]
            key = test[1]
            really_delete = test[2]
            return_value = delete_registry_key(key, really_delete)
            self.assertEqual(rc, return_value)
            if really_delete:
                self.assertFalse(detect_registry_key(key))

        # Test Unicode key.  In BleachBit 0.7.3 this scenario would lead to
        # the error (bug 537109)
        # UnicodeDecodeError: 'ascii' codec can't decode byte 0xc3 in position
        # 11: ordinal not in range(128)
        key = r'Software\\BleachBit\\DeleteThisKey'
        hkey = winreg.CreateKey(
            winreg.HKEY_CURRENT_USER, key + r'\\AndThisKey-Ö')
        hkey.Close()
        return_value = delete_registry_key('HKCU\\' + key, True)
        self.assertTrue(return_value)
        return_value = delete_registry_key('HKCU\\' + key, True)
        self.assertFalse(return_value)

    def test_delete_updates(self):
        """Unit test for delete_updates

        As a preview, this does not modify services or delete files.
        """
        if not shell.IsUserAnAdmin():
            # It should return None without doing any work.
            for _ in delete_updates():
                pass
            return

        counter = 0
        for cmd in delete_updates():
            counter += 1
            self.assertIsInstance(cmd, (Delete, Function))
        logger.debug('delete_updates() returned %s commands', f'{counter:,}')

    def test_is_service_running(self):
        """Unit test for is_service_running()"""
        # RPC is always running.
        self.assertTrue(is_service_running('rpcss'))
        # Windows Update is sometimes running.
        self.assertIsInstance(is_service_running('wuauserv'), bool)
        # Non-existent service should raise an error.
        with self.assertRaises(pywintypes.error):
            is_service_running('does_not_exist')
        # None should raise an error.
        with self.assertRaises(AssertionError):
            is_service_running(None)

    def test_is_ots_elevation_without_flag_returns_false(self):
        """Without --uac-sid-token, is_ots_elevation() returns False"""
        argv = ['bleachbit.exe', '--gui']
        with mock.patch('bleachbit.Windows.sys.argv', argv):
            self.assertFalse(is_ots_elevation())

    def test_is_ots_elevation_false_when_tokens_match(self):
        """If current token matches parent token, there is no elevation"""
        parent_token = 'ABCDEFGH'
        argv = ['bleachbit.exe', '--gui', '--uac-sid-token', parent_token]
        with mock.patch('bleachbit.Windows.sys.argv', argv):
            with mock.patch('bleachbit.Windows.get_sid_token_48', return_value=parent_token):
                self.assertFalse(is_ots_elevation())

    def test_is_ots_elevation_true_when_tokens_differ(self):
        """If current token differs from parent token, elevation is detected"""
        parent_token = 'ABCDEFGH'
        argv = ['bleachbit.exe', '--gui', '--uac-sid-token', parent_token]
        with mock.patch('bleachbit.Windows.sys.argv', argv):
            with mock.patch('bleachbit.Windows.get_sid_token_48', return_value='DIFFERNT'):
                self.assertTrue(is_ots_elevation())

    def test_is_ots_elevation_ignores_flag_without_value(self):
        """A trailing --uac-sid-token without value should not crash and returns False"""
        argv = ['bleachbit.exe', '--gui', '--uac-sid-token']
        with mock.patch('bleachbit.Windows.sys.argv', argv):
            self.assertFalse(is_ots_elevation())

    def test_is_ots_elevation_returns_false_on_get_sid_error(self):
        """If get_sid_token_48() raises, is_ots_elevation() falls back to False"""
        parent_token = 'ABCDEFGH'
        argv = ['bleachbit.exe', '--gui', '--uac-sid-token', parent_token]
        with mock.patch('bleachbit.Windows.sys.argv', argv):
            with mock.patch('bleachbit.Windows.get_sid_token_48', side_effect=RuntimeError('error')):
                self.assertFalse(is_ots_elevation())

    def test_splash_thread_reuses_cached_class_atom(self):
        """_register_window_class skips RegisterClass when cached."""
        splash = SplashThread()
        self.addCleanup(lambda: setattr(SplashThread, '_class_atom', None))
        SplashThread._class_atom = 9876
        with mock.patch('bleachbit.Windows.win32gui.RegisterClass') as mock_register:
            atom = splash._register_window_class(mock.Mock())
        self.assertEqual(atom, 9876)
        mock_register.assert_not_called()

    def test_splash_thread_recovers_when_class_exists(self):
        """_register_window_class handles ERROR_CLASS_ALREADY_EXISTS."""
        splash = SplashThread()
        self.addCleanup(lambda: setattr(SplashThread, '_class_atom', None))
        SplashThread._class_atom = None
        wnd_class = mock.Mock()
        wnd_class.hInstance = mock.sentinel.instance
        wnd_class.lpszClassName = 'SimpleWin32'
        register_error = pywintypes.error(
            1410, 'RegisterClass', 'Class already exists.')
        with mock.patch('bleachbit.Windows.win32gui.RegisterClass', side_effect=register_error) as mock_register, \
                mock.patch('bleachbit.Windows.win32gui.GetClassInfo', return_value=(4321,)) as mock_get_class_info:
            atom = splash._register_window_class(wnd_class)
        self.assertEqual(atom, 4321)
        self.assertEqual(SplashThread._class_atom, 4321)
        mock_register.assert_called_once_with(wnd_class)
        mock_get_class_info.assert_called_once_with(
            wnd_class.hInstance, wnd_class.lpszClassName)

    def test_splash_thread_suppresses_startup_error(self):
        """start() does not fail the GUI if splash initialization fails."""
        splash = SplashThread()
        with mock.patch.object(splash, '_show_splash_screen', side_effect=RuntimeError('boom')):
            splash.start()
        self.assertIsNotNone(splash._startup_error)

    def test_splash_thread_join_handles_missing_window(self):
        """join() succeeds even when window handle was never created."""
        splash = SplashThread()
        splash._startup_error = RuntimeError('boom')
        splash._splash_screen_started.set()
        splash.start = lambda: None  # prevent thread start
        # Directly call join; should not raise even though handle is None
        splash.join(timeout=0)

    def test_splash_thread_join_uses_send_message_timeout(self):
        """join() closes the splash synchronously with a timeout."""
        splash = SplashThread()
        splash._splash_screen_handle = 1234
        with mock.patch.object(splash, 'is_alive', return_value=True), \
                mock.patch.object(splash, '_hide_window') as mock_hide, \
                mock.patch('bleachbit.Windows.win32gui.IsWindow', return_value=True), \
                mock.patch(
                    'bleachbit.Windows.ctypes.windll.user32.SendMessageTimeoutW',
                    return_value=1) as mock_send, \
                mock.patch('bleachbit.Windows.win32gui.PostMessage') as mock_post, \
                mock.patch('bleachbit.Windows.Thread.join') as mock_thread_join:
            splash.join(timeout=0)

        mock_send.assert_called_once()
        mock_hide.assert_called_once_with(1234)
        mock_post.assert_not_called()
        mock_thread_join.assert_called_once_with(splash, timeout=0)

    def test_splash_thread_join_falls_back_after_send_failure(self):
        """join() falls back to PostMessage if synchronous close fails."""
        splash = SplashThread()
        splash._splash_screen_handle = 1234
        with mock.patch.object(splash, 'is_alive', return_value=True), \
                mock.patch.object(splash, '_hide_window'), \
                mock.patch('bleachbit.Windows.win32gui.IsWindow', return_value=True), \
                mock.patch(
                    'bleachbit.Windows.ctypes.windll.user32.SendMessageTimeoutW',
                    side_effect=RuntimeError('boom')), \
                mock.patch('bleachbit.Windows.win32gui.PostMessage') as mock_post, \
                mock.patch('bleachbit.Windows.Thread.join'):
            splash.join(timeout=0)

        mock_post.assert_called_once_with(
            1234, Windows.win32con.WM_CLOSE, 0, 0)

    def test_get_splash_screen_delay_seconds_default(self):
        with mock.patch.dict(os.environ, {}, clear=False):
            os.environ.pop('BLEACHBIT_SPLASH_SCREEN_DELAY', None)
            self.assertEqual(get_splash_screen_delay_seconds(), 0.0)

    def test_get_splash_screen_delay_seconds_valid(self):
        with mock.patch.dict(os.environ, {'BLEACHBIT_SPLASH_SCREEN_DELAY': '3.5'}, clear=False):
            self.assertEqual(get_splash_screen_delay_seconds(), 3.5)

    def test_get_splash_screen_delay_seconds_invalid(self):
        with mock.patch.dict(os.environ, {'BLEACHBIT_SPLASH_SCREEN_DELAY': 'abc'}, clear=False):
            self.assertEqual(get_splash_screen_delay_seconds(), 0.0)

    def test_get_splash_screen_delay_seconds_negative(self):
        with mock.patch.dict(os.environ, {'BLEACHBIT_SPLASH_SCREEN_DELAY': '-1'}, clear=False):
            self.assertEqual(get_splash_screen_delay_seconds(), 0.0)

    @common.skipUnlessDestructive
    def test_run_net_service_command(self):
        """Integration test for run_net_service_command().

        Actually stop/start Windows Update service.

        spooler (Print Spooler) is often on by default and has no dependencies.

        Windows Audio Endpoint Builder (AudioEndpointBuilder) is often on
        by default and depends on audiosrv (Windows Audio).

        Requires admin.
        """
        if not shell.IsUserAnAdmin():
            self.skipTest('requires administrator privileges')

        def _service_exists_and_enabled(svc):
            scm = win32service.OpenSCManager(
                None, None, win32service.SC_MANAGER_CONNECT)
            try:
                hs = win32service.OpenService(
                    scm, svc, win32service.SERVICE_QUERY_STATUS | win32service.SERVICE_QUERY_CONFIG)
                try:
                    cfg = win32service.QueryServiceConfig(hs)
                    return True if cfg[1] != win32service.SERVICE_DISABLED else False
                finally:
                    win32service.CloseServiceHandle(hs)
            except pywintypes.error:
                return False
            finally:
                win32service.CloseServiceHandle(scm)

        def _can_open_all_access(svc):
            scm = win32service.OpenSCManager(
                None, None, win32service.SC_MANAGER_CONNECT)
            try:
                try:
                    hs = win32service.OpenService(
                        scm, svc, win32service.SERVICE_ALL_ACCESS)
                except pywintypes.error:
                    return False
                else:
                    win32service.CloseServiceHandle(hs)
                    return True
            finally:
                win32service.CloseServiceHandle(scm)

        is_ci = os.environ.get('APPVEYOR') == 'True'
        if is_ci:
            candidates = ('bits', 'wuauserv',
                          'AudioEndpointBuilder', 'spooler')
            service = None
            # Prefer already-running to avoid state changes
            for s in candidates:
                if _service_exists_and_enabled(s) and _can_open_all_access(s) and is_service_running(s):
                    service = s
                    break
            # If none are running, pick any we can open with required access
            if not service:
                for s in candidates:
                    if _service_exists_and_enabled(s) and _can_open_all_access(s):
                        service = s
                        break
        else:
            candidates = ('AudioEndpointBuilder',
                          'spooler', 'bits', 'wuauserv')
            service = None
            for s in candidates:
                if _service_exists_and_enabled(s):
                    service = s
                    break
        if not service:
            self.skipTest('no suitable startable service on this machine')

        initial_running = is_service_running(service)

        try:
            if is_ci:
                ret = run_net_service_command(service, True)
                self.assertEqual(ret, 0)
                ret = run_net_service_command(service, True)
                self.assertEqual(ret, 0)
            else:
                ret = run_net_service_command(service, False)
                self.assertEqual(ret, 0)
                self.assertFalse(is_service_running(service))
                if service == 'AudioEndpointBuilder':
                    self.assertFalse(is_service_running('audiosrv'))
                ret = run_net_service_command(service, False)
                self.assertEqual(ret, 0)
                self.assertFalse(is_service_running(service))
                ret = run_net_service_command(service, True)
                self.assertEqual(ret, 0)
                self.assertTrue(is_service_running(service))
                ret = run_net_service_command(service, True)
                self.assertEqual(ret, 0)
                self.assertTrue(is_service_running(service))
        finally:
            if not is_ci:
                run_net_service_command(service, initial_running)
                self.assertEqual(initial_running, is_service_running(service))

    def test_run_net_service_command_not_admin(self):
        """Test as run_net_service_command() as not admin user"""
        if shell.IsUserAnAdmin():
            self.skipTest('requires non-admin user')
        service = 'wuauserv'
        initial_running = is_service_running(service)
        for start in (True, False):
            with self.assertRaises(RuntimeError):
                run_net_service_command(service, start)
            self.assertEqual(is_service_running(service), initial_running)

    def test_run_net_service_command_invalid_service(self):
        """Test as run_net_service_command() with invalid service"""
        for service in ('does_not_exist', None):
            for start in (True, False):
                with self.subTest(service=service, start=start):
                    with self.assertRaises((AssertionError, RuntimeError)):
                        run_net_service_command(service, start)

    def test_delete_registry_value(self):
        """Unit test for delete_registry_value"""

        #
        # test: value does exist
        #

        # create a name-value pair
        key = 'Software\\BleachBit'
        hkey = winreg.CreateKey(winreg.HKEY_CURRENT_USER, key)

        value_name = 'delete_this_value_name'
        winreg.SetValueEx(
            hkey, value_name, 0, winreg.REG_SZ, 'delete this value')
        hkey.Close()

        # delete and confirm
        self.assertTrue(
            delete_registry_value('HKCU\\' + key, value_name, False))
        self.assertTrue(
            delete_registry_value('HKCU\\' + key, value_name, True))
        self.assertFalse(
            delete_registry_value('HKCU\\' + key, value_name, False))
        self.assertFalse(
            delete_registry_value('HKCU\\' + key, value_name, True))

        #
        # test: value does not exist
        #
        self.assertFalse(delete_registry_value(
            'HKCU\\' + key, 'doesnotexist', False))
        self.assertFalse(delete_registry_value(
            'HKCU\\' + key, 'doesnotexist', True))
        self.assertFalse(delete_registry_value(
            'HKCU\\doesnotexist', value_name, False))
        self.assertFalse(delete_registry_value(
            'HKCU\\doesnotexist', value_name, True))

    def test_detect_registry_key(self):
        """Test for detect_registry_key()"""
        self.assertTrue(detect_registry_key('HKCU\\Software\\Microsoft\\'))
        self.assertTrue(not detect_registry_key(
            'HKCU\\Software\\DoesNotExist'))

    def test_get_clipboard_paths(self):
        """Unit test for get_clipboard_paths"""
        # The clipboard is an unknown state, so check the function does
        # not crash and that it returns the right data type.
        paths = get_clipboard_paths()
        self.assertIsInstance(paths, (type(None), tuple))

        # Set the clipboard to an unsupported type (text), so expect no
        # files are returned
        import win32clipboard
        win32clipboard.OpenClipboard()
        win32clipboard.EmptyClipboard()
        fname = r'c:\windows\notepad.exe'
        win32clipboard.SetClipboardText(fname, win32clipboard.CF_TEXT)
        win32clipboard.SetClipboardText(fname, win32clipboard.CF_UNICODETEXT)
        self.assertEqual(win32clipboard.GetClipboardData(
            win32clipboard.CF_TEXT), fname.encode('ascii'))
        self.assertEqual(win32clipboard.GetClipboardData(
            win32clipboard.CF_UNICODETEXT), fname)
        win32clipboard.CloseClipboard()

        paths = get_clipboard_paths()
        self.assertIsInstance(paths, (type(None), tuple))
        self.assertEqual(paths, ())

        # Put files in the clipboard in supported format
        args = ('powershell.exe', 'Set-Clipboard',
                '-Path', r'c:\windows\*.exe')
        (ext_rc, _stdout, _stderr) = General.run_external(args)
        self.assertEqual(ext_rc, 0)
        paths = get_clipboard_paths()
        self.assertIsInstance(paths, (type(None), tuple))
        self.assertGreater(len(paths), 1)
        for path in paths:
            self.assertExists(path)

    def test_get_font_conf_file(self):
        """Unit test for get_font_conf_file"""
        # This tests only one of three situations.
        font_fn = get_font_conf_file()
        self.assertExists(font_fn)

    def test_get_known_folder_path(self):
        """Unit test for get_known_folder_path"""
        ret = get_known_folder_path('LocalAppDataLow')
        self.assertNotEqual(ret, '')
        self.assertNotEqual(ret, None)
        self.assertExists(ret)

    def test_get_fixed_drives(self):
        """Unit test for get_fixed_drives"""
        drives = []
        for drive in get_fixed_drives():
            drives.append(drive)
            self.assertEqual(drive, drive.upper())
        self.assertIn("C:\\", drives)

    def test_get_sid_token_48_basic_properties(self):
        """get_sid_token_48() returns an 8-char, URL-safe ASCII string"""
        token = get_sid_token_48()
        self.assertIsString(token)
        # 6 bytes (48 bits) become 8 base64-url characters without padding
        self.assertEqual(len(token), 8)
        for ch in token:
            self.assertLess(ord(ch), 128)
        # urlsafe_b64encode must not use '+', '/', or '='
        self.assertNotIn('+', token)
        self.assertNotIn('/', token)
        self.assertNotIn('=', token)

    def test_get_sid_token_48_is_deterministic_for_current_process(self):
        """Multiple calls for the same process should yield the same token"""
        t1 = get_sid_token_48()
        t2 = get_sid_token_48()
        self.assertEqual(t1, t2)

    def test_get_windows_version(self):
        """Unit test for get_windows_version"""
        v = get_windows_version()
        self.assertGreaterEqual(v, 5.1)
        self.assertGreater(v, 5)
        self.assertIsInstance(v, Decimal)

    def test_empty_recycle_bin(self):
        """Unit test for empty_recycle_bin"""
        # check the function basically works
        for drive in get_fixed_drives():
            ret = empty_recycle_bin(drive, really_delete=False)
            self.assertIsInteger(ret)

    @common.skipUnlessDestructive
    def test_empty_recycle_bin_per_drive_destructive(self):
        """Empty recycle bin in each drive individually"""
        put_objects_into_recycle_bin()
        for drive in get_fixed_drives():
            with self.subTest(drive=drive):
                try:
                    ret = empty_recycle_bin(drive, really_delete=True)
                except pywintypes.com_error as e:
                    if e.args[0] == -2147024893 and 'APPVEYOR' in os.environ:
                        self.skipTest(
                            'reproducible only under AppVeyor and does not '
                            'test a scenario used outside the tests')
                    raise
                self.assertIsInteger(ret)

    @common.skipUnlessDestructive
    def test_empty_recycle_bin_all_drives_destructive(self):
        """Empty recycle bin in all drives at once"""
        put_objects_into_recycle_bin()
        ret = empty_recycle_bin(None, really_delete=True)
        self.assertIsInteger(ret)

        # Verify that there are no objects in the bin.
        for _f in get_recycle_bin():
            self.fail('recycle bin should be empty, but it is not')

        # Repeat the call to empty for two reasons.
        # 1. Trying to empty an empty recycling bin can cause
        #    a 'catastrophic failure' error (handled in the function)
        # 2. It should show zero bytes were deleted
        for drive in get_fixed_drives():
            ret = empty_recycle_bin(drive, really_delete=True)
            self.assertEqual(ret, 0)

    def test_file_wipe(self):
        """Unit test for file_wipe

        There are more tests in testwipe.py
        """

        from bleachbit.WindowsWipe import file_wipe, open_file, close_file, file_make_sparse
        from bleachbit.Windows import elevate_privileges
        from win32con import GENERIC_WRITE, WRITE_DAC

        dirname = self.mkdtemp()

        filenames = ('short', 'long' + 'x' * 250, 'utf8-ɡælɪk')
        for filename in filenames:
            longname = os.path.join(dirname, filename)
            logger.debug('file_wipe(%s)', longname)

            def _write_file(longname, contents):
                self.write_file(longname, contents)
                import win32api
                shortname = extended_path_undo(
                    win32api.GetShortPathName(extended_path(longname)))
                self.assertExists(shortname)
                return shortname

            def _deny_access(fh):
                import win32security
                import ntsecuritycon as con

                user, _, _ = win32security.LookupAccountName(
                    "", win32api.GetUserName())
                dacl = win32security.ACL()
                dacl.AddAccessDeniedAce(
                    win32security.ACL_REVISION, con.FILE_GENERIC_READ | con.FILE_GENERIC_WRITE, user)
                win32security.SetSecurityInfo(fh, win32security.SE_FILE_OBJECT, win32security.DACL_SECURITY_INFORMATION,
                                              None, None, dacl, None)

            def _test_wipe(contents, deny_access=False, is_sparse=False):
                shortname = _write_file(longname, contents)
                if deny_access or is_sparse:
                    fh = open_file(extended_path(longname),
                                   mode=GENERIC_WRITE | WRITE_DAC)
                    if is_sparse:
                        file_make_sparse(fh)
                    if deny_access:
                        _deny_access(fh)
                    close_file(fh)
                logger.debug('test_file_wipe(): filename length={}, shortname length ={}, contents length={}, is_sparse={}'.format(
                    len(longname), len(shortname), len(contents), is_sparse))
                if shell.IsUserAnAdmin():
                    # wiping requires admin privileges
                    file_wipe(shortname)
                    file_wipe(longname)
                else:
                    with self.assertRaises(pywintypes.error):
                        file_wipe(shortname)
                        file_wipe(longname)
                self.assertExists(shortname)
                os.remove(extended_path(shortname))
                self.assertNotExists(shortname)

            # A small file that fits in MFT
            _test_wipe(b'')

            # requires wiping of extents
            _test_wipe(b'secret' * 100000)

            # requires wiping of extents: special file case
            elevate_privileges(False)
            _test_wipe(b'secret' * 100000, deny_access=True, is_sparse=True)

        shutil.rmtree(dirname, True)

        if shell.IsUserAnAdmin():
            logger.warning(
                'You should also run test_file_wipe() without admin privileges.')
        else:
            logger.warning(
                'You should also run test_file_wipe() with admin privileges.')

    def test_setup_environment(self):
        """Unit test for setup_environment"""
        setup_environment()
        envs = ['commonappdata', 'documents', 'music', 'pictures', 'video',
                'localappdata', 'localappdatalow']
        for env in envs:
            self.assertExists(os.environ[env])

    def test_split_registry_key(self):
        """Unit test for split_registry_key"""
        tests = (('HKCU\\Software', winreg.HKEY_CURRENT_USER, 'Software'),
                 ('HKLM\\SOFTWARE', winreg.HKEY_LOCAL_MACHINE, 'SOFTWARE'),
                 ('HKU\\.DEFAULT', winreg.HKEY_USERS, '.DEFAULT'))
        for (input_key, expected_hive, expected_key) in tests:
            (hive, key) = split_registry_key(input_key)
            self.assertEqual(expected_hive, hive)
            self.assertEqual(expected_key, key)

    def test_read_registry_key(self):
        """Unit test for read_registry_key"""
        tests = (('HKCR\\.bmp', 'PerceivedType', 'image'),
                 ('HKCU\\Software\\BleachBit\\DoesNotExist', 'DoesNotExist', None))
        for (input_key, input_value, expected_value) in tests:
            value = read_registry_key(input_key, input_value)
            if value != None:
                value = value.lower()  # AppVeyor: image, Windows 11: Image
            self.assertEqual(expected_value, value)

    def test_parse_windows_build(self):
        """Unit test for parse_windows_build"""
        tests = (('5.1.2600', Decimal('5.1')),
                 ('5.1', Decimal('5.1')),
                 ('10.0.10240', 10),
                 ('10.0', 10))
        for test in tests:
            self.assertEqual(parse_windows_build(test[0]), test[1])

        # test for crash
        parse_windows_build()
        parse_windows_build(platform.version())
        parse_windows_build(platform.uname()[3])

    def test_path_on_network(self):
        """Unit test for path_on_network"""
        self.assertFalse(path_on_network('c:\\bleachbit.exe'))
        self.assertFalse(path_on_network('a:\\bleachbit.exe'))
        self.assertTrue(path_on_network('\\\\Server\\Folder\\bleachbit.exe'))

    def test_shell_change_notify(self):
        """Unit test for shell_change_notify"""
        ret = shell_change_notify()
        self.assertEqual(ret, 0)

    def test_splash_screen(self):
        """Unit test for splash screen"""
        splash_thread = SplashThread()
        icon_path = splash_thread.get_icon_path()
        self.assertTrue(icon_path.is_absolute())
        self.assertExists(icon_path)
        splash_thread.start()
        timeout = 5.0  # seconds
        start_time = time.time()
        time.sleep(1)
        while not splash_thread.is_alive() and (time.time() - start_time) < timeout:
            time.sleep(0.1)
        self.assertTrue(splash_thread.is_alive(),
                        'Splash thread did not start within timeout')
        splash_thread.join()

        # Originally, running twice in same process caused indefinite hang,
        # so run it twice.
        repeat = SplashThread()
        repeat.start()
        repeat.join()

    def test_splash_screen_window_pos(self):
        """Unit test for calculate_window_position()"""
        splash_thread = SplashThread()

        # Test with a typical 1080p display
        display_width = 1920
        display_height = 1080
        x, y, width, height = splash_thread.calculate_window_position(
            display_width, display_height)

        # The window should have positive dimensions
        self.assertGreater(width, 0)
        self.assertGreater(height, 0)

        # The window position should be positive (on screen)
        self.assertGreaterEqual(x, 0)
        self.assertGreaterEqual(y, 0)

        # The window should fit within the display
        self.assertLessEqual(x + width, display_width)
        self.assertLessEqual(y + height, display_height)

        # The window should be roughly centered
        # x should be approximately (display_width - width) / 2
        expected_x = (display_width - width) // 2
        self.assertEqual(x, expected_x)
        # y should be approximately (display_height - height) / 2
        expected_y = (display_height - height) // 2
        self.assertEqual(y, expected_y)

        # Splash screen is the square logo.
        self.assertEqual(width, 256)
        self.assertEqual(height, 256)

    def test_set_environ(self):
        for folder in ['folderäö', 'folder']:
            test_dir = os.path.join(self.tempdir, folder)
            os.mkdir(test_dir)
            self.assertExists(test_dir)
            set_environ('cd_test', test_dir)
            self.assertEqual(os.environ['cd_test'], test_dir)
            os.environ.pop('cd_test')



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.

"""
Build BleachBit tarballs
"""

# standard library
import glob
import os
import sys

# local import
import bleachbit


APP_DESCRIPTION = "BleachBit frees space and maintains privacy by quickly wiping files you don't need and didn't know you had."

data_files = []
if sys.platform == 'linux':
    data_files.append(('/usr/share/applications',
                      ['./org.bleachbit.BleachBit.desktop']))
    data_files.append(('/usr/share/pixmaps/', ['./bleachbit.png']))
elif sys.platform[:6] == 'netbsd':
    data_files.append(('/usr/pkg/share/applications',
                      ['./org.bleachbit.BleachBit.desktop']))
    data_files.append(('/usr/pkg/share/pixmaps/', ['./bleachbit.png']))
elif sys.platform.startswith('openbsd') or sys.platform.startswith('freebsd'):
    data_files.append(
        ('/usr/local/share/applications', ['./org.bleachbit.BleachBit.desktop']))
    data_files.append(('/usr/local/share/pixmaps/', ['./bleachbit.png']))


def run_setup(args=None):
    """Run setup from setuptools"""
    if args is None:
        args = {}
    # pylint: disable=import-outside-toplevel
    from setuptools import setup
    setup(name='bleachbit',
          version=bleachbit.APP_VERSION,
          description=bleachbit.APP_NAME,
          long_description=APP_DESCRIPTION,
          author="Andrew Ziem",
          author_email="andrew@bleachbit.org",
          url=bleachbit.APP_URL,
          download_url="https://www.bleachbit.org/download",
          classifiers=[
              'Development Status :: 5 - Production/Stable',
              'Programming Language :: Python',
              'Operating System :: Microsoft :: Windows',
              'Operating System :: POSIX :: Linux',
          ],
          license='GPL-3.0-or-later',
          # py_requires='>=3.8',
          # Ubuntu 20.04 LTS is EOL on April 2025, and it has Python 3.8.
          platforms='Linux and Windows, Python v3.8+, GTK v3.24+',
          packages=['bleachbit', 'bleachbit.markovify'],
          **args)


def supported_languages():
    """Return list of supported languages by scanning ./po/

    This function is used by:
    * windows/setup.py
    * bleachbit-misc/extract_desktop.py
    """
    langs = []
    for pathname in glob.glob('po/*.po'):
        basename = os.path.basename(pathname)
        langs.append(os.path.splitext(basename)[0])
    return sorted(langs)


if __name__ == '__main__':
    run_setup()



# vim: ts=4:sw=4:expandtab
# -*- coding: UTF-8 -*-

# BleachBit
# Copyright (C) 2008-2025 Andrew Ziem
# https://www.bleachbit.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Test case for module Chaff
"""

from unittest import mock
import os
from tempfile import mkdtemp
from shutil import rmtree

from tests import common
from bleachbit.Chaff import download_models, generate_emails, generate_2600, have_models, MODEL_BASENAMES, DEFAULT_MODELS_DIR
from bleachbit.FileUtilities import getsize


class ChaffTestCase(common.BleachbitTestCase):
    """Test case for module Chaff"""

    def test_Chaff(self):
        """Unit test for class Chaff"""
        tmp_dir = mkdtemp(prefix='bleachbit-chaff')
        models_dir = os.path.join(tmp_dir, 'model')
        os.mkdir(models_dir)
        # con=Clinton content
        con_path = os.path.join(models_dir, 'clinton_content_model.json.bz2')
        # sub=Clinton subject
        sub_path = os.path.join(models_dir, 'clinton_subject_model.json.bz2')
        # ts = 2600 Magazine
        ts_path = os.path.join(models_dir, '2600_model.json.bz2')

        for _i in ('download', 'already downloaded'):
            ret = download_models(models_dir=models_dir)
            self.assertIsInstance(ret, bool)
            self.assertTrue(
                ret, "download_models() return False, meaning a download error occurred")

        self.assertExists(con_path)
        self.assertExists(sub_path)
        self.assertExists(ts_path)

        generated_file_names = generate_emails(5, tmp_dir, models_dir)

        self.assertEqual(len(generated_file_names), 5)

        for fn in generated_file_names:
            self.assertExists(fn)
            self.assertGreater(getsize(fn), 100)
            with open(fn) as f:
                contents = f.read()
                self.assertIn('To: ', contents)
                self.assertIn('From: ', contents)
                self.assertIn('Sent: ', contents)
                self.assertIn('Subject: ', contents)
                self.assertNotIn('base64', contents)

        generated_file_names = generate_2600(5, tmp_dir, models_dir)

        rmtree(tmp_dir)

    @mock.patch('bleachbit.Network.download_url_to_fn')
    def test_download_models_fallback(self, mock_download):
        """Test the fallback mechanism in download_models()"""
        tmp_dir = mkdtemp(prefix='bleachbit-chaff')

        # Test when primary download mirror fails but secondary succeeds.
        def succeed_on_second(*args, **_kwargs):
            url = args[0]
            if 'sourceforge' in url:
                return False
            if 'bleachbit.org' in url:
                return True
        mock_download.side_effect = succeed_on_second
        ret = download_models(models_dir=tmp_dir)
        self.assertTrue(
            ret, "download_models() return False, meaning a download error occurred")
        self.assertEqual(mock_download.call_count, 6)

        # Test when both primary and secondary download mirrors will fail.
        mock_download.reset_mock()
        mock_download.side_effect = None
        mock_download.return_value = False
        ret = download_models(models_dir=tmp_dir)
        self.assertFalse(
            ret, "download_models() should return False when any download fails")
        self.assertEqual(mock_download.call_count, 2)

        rmtree(tmp_dir)

    def test_make_sentence_none(self):
        """Test that make_sentence() returning None does not cause TypeError"""
        from bleachbit.Chaff import _get_random_content, _generate_2600_file, _generate_email

        model = mock.MagicMock()
        model.make_sentence.return_value = None
        model.make_short_sentence.return_value = None

        result = _get_random_content(model, number_of_sentences=2)
        self.assertIsNotNone(result)

        result = _generate_2600_file(model, number_of_sentences=2)
        self.assertIsInstance(result, str)

        subject_model = mock.MagicMock()
        subject_model.make_short_sentence.return_value = None
        msg = _generate_email(subject_model, model, number_of_sentences=2)
        self.assertEqual(msg['Subject'], '')

    def test_have_models(self):
        """Test for function have_models()"""
        download_models()
        rc = have_models()
        self.assertTrue(
            rc, "have_models() should return True when models are present")
        self.assertIsInstance(rc, bool)
        for basename in MODEL_BASENAMES:
            fn = os.path.join(DEFAULT_MODELS_DIR, basename)
            self.assertExists(fn)
            os.remove(fn)
        self.assertFalse(
            have_models(), "have_models() should return False when any model is missing")



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.


"""
Test case for module GtkShim
"""

import ctypes
import os
import subprocess
import sys
import unittest
from unittest import mock

from tests import common
from bleachbit import IS_WINDOWS
from bleachbit.General import get_executable
from bleachbit.GtkShim import (
    _build_error_html,
    _fix_arg,
    _handle_gtk_import_error,
    _patched_argv,
    _show_windows_error_dialog,
    _try_import_gtk,
    path_has_lib_or_bin,
)


class PathHasLibOrBinTestCase(unittest.TestCase):
    """Test path_has_lib_or_bin() detection logic."""

    def test_path_has_lib_or_bin(self):
        """Test path_has_lib_or_bin() with various paths."""
        tests = [
            # (path, expected_result)
            # Standalone 'bin' directory triggers detection
            (r'C:\bin', 'bin'),
            (r'C:\bin\xlib', 'bin'),
            (r'C:\libx\bin', 'bin'),
            (r'C:\bin\bin', 'bin'),
            (r'C:\Users\username\Downloads\bin\BleachBit-Portable', 'bin'),
            # Standalone 'lib' directory triggers detection
            (r'C:\lib', 'lib'),
            (r'C:\lib\xbin', 'lib'),
            (r'C:\binx\lib', 'lib'),
            (r'C:\lib\lib', 'lib'),
            (r'C:\Users\username\Downloads\lib\BleachBit-Portable', 'lib'),
            (r'C:\ProgramData\chocolatey\lib\bleachbit.portable\BleachBit-Portable', 'lib'),
            # Detection is case-insensitive
            (r'C:\Users\BIN\app', 'bin'),
            (r'C:\Users\Lib\app', 'lib'),
            # 'binx' is not 'bin'
            (r'C:\Users\username\Downloads\binx\BleachBit-Portable', None),
            # 'xbin' is not 'bin'
            (r'C:\Users\username\Downloads\xbin\BleachBit-Portable', None),
            # 'libfoo' is not 'lib'
            (r'C:\ProgramData\libfoo\app', None),
            # Normal portable path without lib or bin
            (r'C:\Users\username\Downloads\BleachBit-5.0.0-portable\BleachBit-Portable', None),
            # Empty path returns None
            ('', None),
            # 'lib' or 'bin' embedded in longer names should not match
            (r'C:\calibre\library\app', None),
            (r'C:\cabinet\app', None),
        ]
        case_functions = (
            lambda x: x.lower(),
            lambda x: x.upper(),
            lambda x: x.title(),
            lambda x: x.title().swapcase(),
        )
        for path, expected in tests:
            with self.subTest(path=path, expected=expected):
                joined = os.path.join(path, 'bleachbit.exe')
                for func in case_functions:
                    self.assertEqual(path_has_lib_or_bin(
                        func(joined)), expected)


class BuildErrorHtmlTestCase(unittest.TestCase):
    """Tests for _build_error_html()."""

    def test_unknown_error_includes_traceback_and_sysinfo(self):
        """Unknown-error HTML must include traceback and system information."""
        html = _build_error_html(
            RuntimeError('something broke'),
            traceback_text='Traceback (most recent call last):...',
        )
        self.assertIn('Error: something broke', html)
        self.assertIn('Traceback (most recent call last):...', html)
        self.assertIn('System information:\n', html)
        self.assertIn('textarea', html)
        self.assertIn('Copy to clipboard', html)
        self.assertIn("function copyBugReport()", html)
        self.assertIn('get-help', html)
        self.assertEqual(html.count('<textarea'), 1)
        self.assertEqual(html.count('<pre>'), 0)

    def test_unknown_error_escapes_html_in_error_message(self):
        """Error message containing HTML special chars must be escaped."""
        html = _build_error_html(
            RuntimeError('<script>alert(1)</script>'),
        )
        self.assertNotIn('<script>', html)
        self.assertIn('&lt;script&gt;', html)

    def test_html_structure(self):
        """Output must be valid minimal HTML with required elements."""
        html = _build_error_html(
            Exception('test'),
        )
        self.assertIn('<!DOCTYPE html>', html)
        self.assertIn('<html', html)
        self.assertIn('BleachBit cannot start', html)


class ShowWindowsErrorDialogTestCase(unittest.TestCase):
    """Tests for _show_windows_error_dialog() (non-Windows stubs)."""

    @common.skipUnlessWindows
    def test_yes_writes_html_file(self):
        """Selecting Yes must write the HTML file and open the browser."""
        with mock.patch.object(
            ctypes.windll.user32, 'MessageBoxW', return_value=6
        ), mock.patch('webbrowser.open') as mock_open:
            _show_windows_error_dialog('BleachBit', '<html>test</html>')
            mock_open.assert_called_once()
            html_path = mock_open.call_args[0][0].replace('file:///', '')
            with open(html_path, encoding='utf-8') as f:
                self.assertIn('test', f.read())

    @common.skipUnlessWindows
    def test_no_does_not_write_file(self):
        """Selecting No must not open the browser."""
        with mock.patch.object(
            ctypes.windll.user32, 'MessageBoxW', return_value=7
        ), mock.patch('webbrowser.open') as mock_open:
            _show_windows_error_dialog('BleachBit', '<html/>')
            mock_open.assert_not_called()


class HandleGtkImportErrorTestCase(unittest.TestCase):
    """Tests for _handle_gtk_import_error()."""

    @common.skipIfWindows
    def test_noop_on_non_windows(self):
        """Must do nothing on non-Windows platforms."""
        with mock.patch('bleachbit.GtkShim._show_windows_error_dialog') as m:
            _handle_gtk_import_error(ValueError('test'))
            m.assert_not_called()

    @common.skipUnlessWindows
    def test_unknown_error_includes_traceback_in_html(self):
        """Unknown errors must produce HTML with traceback section."""
        with mock.patch('bleachbit.GtkShim._show_windows_error_dialog') as m, \
                mock.patch('sys.frozen', True, create=True), \
                mock.patch('sys.executable', r'C:\Users\user\Apps\BleachBit\bleachbit.exe'):
            try:
                raise RuntimeError('something broke')
            except RuntimeError as e:
                _handle_gtk_import_error(e)
            m.assert_called_once()
            _title, html = m.call_args[0]
            self.assertIn('Traceback', html)
            self.assertIn('System information', html)


class FixArgTestCase(unittest.TestCase):
    """Tests for _fix_arg()."""

    def test_ascii_unchanged(self):
        """ASCII strings must pass through unchanged."""
        self.assertEqual(_fix_arg('hello'), 'hello')

    def test_surrogate_replaced(self):
        """Unpaired surrogates must be replaced with the replacement character."""
        arg = 'prefix' + chr(0xD800) + 'suffix'
        fixed = _fix_arg(arg)
        self.assertNotIn(chr(0xD800), fixed)
        self.assertIn('\ufffd', fixed)

    def test_valid_unicode_unchanged(self):
        """Valid Unicode characters must pass through unchanged."""
        self.assertEqual(_fix_arg('\u2014em-dash'), '\u2014em-dash')
        self.assertEqual(_fix_arg('\U0001F525'), '\U0001F525')


class PatchedArgvTestCase(unittest.TestCase):
    """Tests for _patched_argv()."""

    def test_transform_applied(self):
        """Transform function must be applied to all argv elements."""
        original = sys.argv[:]
        with _patched_argv(lambda a: a.swapcase()):
            self.assertEqual(sys.argv, [a.swapcase() for a in original])
        self.assertEqual(sys.argv, original)

    def test_restored_on_exception(self):
        """sys.argv must be restored even when exception occurs."""
        original = sys.argv[:]
        with self.assertRaises(RuntimeError):
            with _patched_argv(lambda a: 'x'):
                raise RuntimeError('boom')
        self.assertEqual(sys.argv, original)

    def test_restores_reference_identity(self):
        """sys.argv must be restored to the exact same object reference."""
        original = sys.argv
        with _patched_argv(lambda a: a):
            pass
        self.assertIs(sys.argv, original)


class TryImportGtkTestCase(unittest.TestCase):
    """Test _try_import_gtk."""

    def test_surrogate_in_argv(self):
        """Must not crash for surrogates in sys.argv."""
        bad_path = os.path.join(
            'c:\\', 'temp', 'surrogate') + chr(0xD800) + 'beta_test0'

        if IS_WINDOWS:
            script = f'''
import ctypes
import sys
from unittest import mock
sys.argv = ['bleachbit.py', {bad_path!r}]
with mock.patch.object(ctypes.windll.user32, 'MessageBoxW', return_value=7):
    from bleachbit.GtkShim import _try_import_gtk
    success, reason = _try_import_gtk()
print(f"OK success={{success}} reason={{reason}}")
'''
        else:
            script = f'''
import sys
sys.argv = ['bleachbit.py', {bad_path!r}]
from bleachbit.GtkShim import _try_import_gtk
success, reason = _try_import_gtk()
print(f"OK success={{success}} reason={{reason}}")
'''

        result = subprocess.run(
            [get_executable(), '-c', script],
            capture_output=True, text=True,
            timeout=30,
        )

        self.assertEqual(result.returncode, 0,
                         f'stderr: {result.stderr}')
        self.assertIn('OK success=', result.stdout,
                      'surrogate argv caused GTK import to crash')


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.


"""
Test case for Common
"""

# standard imports
import os

# first party imports
from tests import common


class CommonTestCase(common.BleachbitTestCase):
    """Test case for Common."""

    _valid_language_codes = [
        'be@latin',
        'C.UTF-8',
        'C.utf8',
        'C',
        'de_DE.iso88591',
        'de-CH',  # seen in Fedora in Docker
        'en_US',
        'en-US',
        'en',
        'es_419',
        'fr_FR.utf8',
        'ja_JP.SJIS',
        'ko_KR.eucKR',
        'nb_NO.ISO-8859-1',
        'POSIX',
        'ru_RU.KOI8-R',
        'zh_Hant',
        'be_BY.utf8@latin',

    ]

    _invalid_language_codes = ['e', 'en_', 'english', 'en_US_', '123',
                               'en_us', 'en_US.',
                               'en_us.utf8',
                               'en_us.UTF-8',
                               'utf8',
                               'UTF-8',
                               '.utf8',
                               '.UTF-8',
                               'en@boldquot.header',
                               'tor@default.service',
                               '',
                               [],
                               0,
                               None]

    def test_assertIsLanguageCode_hardcoded_valid(self):
        """Test assertIsLanguageCode() accepts hard-coded valid codes"""
        for code in self._valid_language_codes:
            with self.subTest(code=code):
                self.assertIsLanguageCode(code)

    def test_assertIsLanguageCode_hardcoded_invalid(self):
        """Test assertIsLanguageCode() rejects hard-coded invalid codes"""
        invalid_codes = list(self._invalid_language_codes)
        invalid_codes.extend([code + ' ' for code in self._valid_language_codes])
        invalid_codes.extend([' ' + code for code in self._valid_language_codes])
        for code in invalid_codes:
            with self.subTest(code=code):
                with self.assertRaises(AssertionError, msg=f'Expected exception for {code}'):
                    self.assertIsLanguageCode(code)

    def test_assertIsLanguageCode_live(self):
        """Test assertIsLanguageCode() using live data"""
        from bleachbit import locale_dir
        locale_dirs = list(set([locale_dir, '/usr/share/locale']))
        lang_codes = []
        # Skip directories that are not valid language codes
        skip_dirs = {'l10n'}
        for locale_dir in locale_dirs:
            if not os.path.isdir(locale_dir):
                continue
            for lang_code in os.listdir(locale_dir):
                if not os.path.isdir(os.path.join(locale_dir, lang_code)):
                    continue
                if lang_code in skip_dirs:
                    continue
                lang_codes.append(lang_code)
        if os.path.exists('/etc/locale.alias'):
            with open('/etc/locale.alias', 'r') as f:
                for line in f:
                    line = line.strip()
                    if not line.startswith('#'):
                        parts = line.split()
                        if len(parts) > 1:
                            lang_codes.append(parts[1])
        for lang_code in lang_codes:
            if lang_code in common.SKIP_ALIAS_CODES:
                continue
            self.assertIsLanguageCode(lang_code)

    def test_environment(self):
        """Test for important environment variables"""
        # useful for researching
        # grep -Poh "([\\$%]\w+)" cleaners/*xml | cut -b2- | sort | uniq -i
        # bleachbit/init.py sets these variables on posix, if they do not exist.
        # Docker does not set USER
        envs = {'posix': ['HOME', 'XDG_CACHE_HOME', 'XDG_CONFIG_HOME', 'XDG_DATA_HOME'],
                'nt': ['AppData', 'CommonAppData', 'Documents', 'ProgramFiles', 'UserProfile', 'WinDir']}
        for env in envs[os.name]:
            e = os.getenv(env)
            self.assertIsNotNone(e, f"Environment variable {env} is not set")
            self.assertNotEqual(
                e.strip(), "", f"Environment variable {env} is empty")

    def test_get_put_env(self):
        """Unit test for get_env() and put_env()"""
        self.assertIsNone(common.get_env('PUTENV_TEST'))

        common.put_env('PUTENV_TEST', '1')
        self.assertEqual('1', common.get_env('PUTENV_TEST'))

        common.put_env('PUTENV_TEST', None)
        self.assertIsNone(common.get_env('PUTENV_TEST'))

    @common.skipUnlessWindows
    def test_get_opened_windows_titles(self):
        """Unit test for get_opened_windows_titles()"""
        titles = common.get_opened_windows_titles()

        self.assertIsInstance(titles, list, "Should return a list")
        self.assertGreater(len(titles), 0,
                           f"Should have at least one window title. Got: {titles}")

        # Verify all items are strings and none are empty
        for title in titles:
            self.assertIsInstance(
                title, str, f"All titles should be strings, got {type(title)}")
            self.assertGreater(len(title.strip()), 0,
                               f"Window title should not be empty or whitespace-only: '{title}'")

    def test_mock_missing_package(self):
        """Test mock_missing_package context manager"""
        # Inside the context, importing logging should raise an errorl.
        with common.mock_missing_package('logging', clear_prefixes=('bleachbit.CLI',)):
            with self.assertRaises(ImportError):
                import logging  # pylint: disable=import-outside-toplevel
            with self.assertRaises(ImportError):
                import bleachbit.CLI  # pylint: disable=import-outside-toplevel
            # Importing sys should be unaffected.
            import sys  # pylint: disable=import-outside-toplevel
        # Outside the context, importing should work.
        import logging  # pylint: disable=import-outside-toplevel
        import bleachbit.CLI  # pylint: disable=import-outside-toplevel

    def test_set_temporary_env(self):
        """Test set_temporary_env context manager"""
        # Test with an environment variable that was not set before.
        self.assertIsNone(os.environ.get('TEST_VAR_NEW'))
        with common.set_temporary_env('TEST_VAR_NEW', 'test_value'):
            self.assertEqual('test_value', os.environ.get('TEST_VAR_NEW'))
        # Outside the context, the environment variable should be unset.
        self.assertIsNone(os.environ.get('TEST_VAR_NEW'))

        # Test with an environment variable that was set before.
        os.environ['TEST_VAR_EXISTING'] = 'original_value'
        with common.set_temporary_env('TEST_VAR_EXISTING', 'override_value'):
            self.assertEqual('override_value',
                             os.environ.get('TEST_VAR_EXISTING'))
        # Outside the context, the environment variable should be restored.
        self.assertEqual('original_value', os.environ.get('TEST_VAR_EXISTING'))
        del os.environ['TEST_VAR_EXISTING']

    def test_touch_file(self):
        """Unit test for touch_file"""
        fn = os.path.join(self.tempdir, 'test_touch_file')
        self.assertNotExists(fn)

        # Create empty file.
        common.touch_file(fn)
        from bleachbit.FileUtilities import getsize
        self.assertExists(fn)
        self.assertEqual(0, getsize(fn))

        # Increase size of file.
        fsize = 2**13
        with open(fn, "w") as f:
            f.write(' ' * fsize)
        self.assertEqual(fsize, getsize(fn))

        # Do not truncate.
        common.touch_file(fn)
        self.assertExists(fn)
        self.assertEqual(fsize, getsize(fn))

    def test_assertExists_relative_path(self):
        """Unit test for assertExists with relative path"""
        # Create a file in the current directory
        os.chdir(self.tempdir)
        relative_fn = 'test_relative_file.txt'
        common.touch_file(relative_fn)
        self.assertExists(relative_fn)
        os.remove(relative_fn)
        self.assertNotExists(relative_fn)
        self.assertNotExists('this-does-not-exist')
